import warnings
from copy import copy as std_copy

import numpy as np

import nengo  # pylint: disable=cyclic-import
from nengo.config import SupportDefaultsMixin
from nengo.exceptions import NotAddedToNetworkWarning, ValidationError
from nengo.params import (
    FrozenObject,
    IntParam,
    NumberParam,
    Parameter,
    StringParam,
    Unconfigurable,
    iter_params,
)
from nengo.utils.numpy import as_shape, is_integer, maxint, maxseed


class NetworkMember(type):
    """A metaclass used to add instances of derived classes to networks.

    Inheriting from this class means that Network.add will be invoked after
    initializing the object, unless add_to_container=False is passed to the
    derived class constructor.
    """

    def __call__(cls, *args, **kwargs):
        """Override default __call__ behavior so that Network.add is called."""
        inst = cls.__new__(cls)
        add_to_container = kwargs.pop("add_to_container", True)
        # Do the __init__ before adding in case __init__ errors out
        inst.__init__(*args, **kwargs)
        if add_to_container:
            nengo.Network.add(inst)
        inst._initialized = True
        return inst


class NengoObject(SupportDefaultsMixin, metaclass=NetworkMember):
    """A base class for Nengo objects.

    Parameters
    ----------
    label : string
        A descriptive label for the object.
    seed : int
        The seed used for random number generation.

    Attributes
    ----------
    label : string
        A descriptive label for the object.
    seed : int
        The seed used for random number generation.
    """

    # Order in which parameters have to be initialized.
    # Missing parameters will be initialized last in an undefined order.
    # This is needed for pickling and copying of Nengo objects when the
    # parameter initialization order matters.
    _param_init_order = []

    label = StringParam("label", default=None, optional=True)
    seed = IntParam("seed", default=None, low=0, high=maxseed, optional=True)

    def __init__(self, label, seed):
        super().__init__()
        self._initialized = False
        self.label = label
        self.seed = seed

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_initialized"] = False

        for attr in self.params:
            param = getattr(type(self), attr)
            if self in param:
                state[attr] = getattr(self, attr)

        return state

    def __setstate__(self, state):
        for attr in self._param_init_order:
            setattr(self, attr, state.pop(attr))

        for attr in self.params:
            if attr in state:
                setattr(self, attr, state.pop(attr))

        for k, v in state.items():
            setattr(self, k, v)

        self._initialized = True
        if len(nengo.Network.context) > 0:
            warnings.warn(NotAddedToNetworkWarning(self))

    def __setattr__(self, name, val):
        initialized = hasattr(self, "_initialized") and self._initialized
        if initialized and not hasattr(self, name):
            warnings.warn(
                f"Creating new attribute '{name}' on '{self}'. "
                "Did you mean to change an existing attribute?",
                SyntaxWarning,
            )
        super().__setattr__(name, val)

    def __str__(self):
        return self._str(include_id=not hasattr(self, "label") or self.label is None)

    def __repr__(self):
        return self._str(include_id=True)

    def _str(self, include_id):
        label = (
            ""
            if not hasattr(self, "label")
            else " (unlabeled)"
            if self.label is None
            else f" '{self.label}'"
        )
        return (
            f"<{type(self).__name__}"
            f"{label}"
            f"{f' at 0x{id(self):x}' if include_id else ''}>"
        )

    @property
    def params(self):
        """Returns a list of parameter names that can be set."""
        return list(iter_params(self))

    def copy(self, add_to_container=True):
        with warnings.catch_warnings():
            # We warn when copying since we can't change add_to_container.
            # However, we deal with it here, so we ignore the warning.
            warnings.simplefilter("ignore", category=NotAddedToNetworkWarning)
            c = std_copy(self)
        if add_to_container:
            nengo.Network.add(c)
        return c


class ObjView:
    """Container for a slice with respect to some object.

    This is used by the __getitem__ of Neurons, Node, and Ensemble, in order
    to pass slices of those objects to Connection. This is a notational
    convenience for creating transforms. See Connection for details.

    Does not currently support any other view-like operations.
    """

    def __init__(self, obj, key=slice(None)):
        self.obj = obj

        # Node.size_in != size_out, so one of these can be invalid
        # NumPy <= 1.8 raises a ValueError instead of an IndexError.
        try:
            self.size_in = np.arange(self.obj.size_in)[key].size
        except (IndexError, ValueError):
            self.size_in = None
        try:
            self.size_out = np.arange(self.obj.size_out)[key].size
        except (IndexError, ValueError):
            self.size_out = None
        if self.size_in is None and self.size_out is None:
            raise IndexError(f"Invalid slice '{key}' of {self.obj}")

        if is_integer(key):
            # single slices of the form [i] should be cast into
            # slice objects for convenience
            if key == -1:
                # special case because slice(-1, 0) gives the empty list
                self.slice = slice(key, None)
            else:
                self.slice = slice(key, key + 1)
        else:
            self.slice = key

    def copy(self):
        return std_copy(self)

    def __len__(self):
        return self.size_out

    def __str__(self):
        return f"{self.obj}[{self._slice_string}]"

    def __repr__(self):
        return f"{self.obj!r}[{self._slice_string}]"

    @property
    def _slice_string(self):
        if isinstance(self.slice, slice):
            sl_start = "" if self.slice.start is None else self.slice.start
            sl_stop = "" if self.slice.stop is None else self.slice.stop
            if self.slice.step is None:
                return f"{sl_start}:{sl_stop}"
            else:
                return f"{sl_start}:{sl_stop}:{self.slice.step}"
        else:
            return str(self.slice)


class NengoObjectParam(Parameter):
    def __init__(
        self,
        name,
        optional=False,
        readonly=True,
        nonzero_size_in=False,
        nonzero_size_out=False,
    ):
        default = Unconfigurable  # These can't have defaults
        self.nonzero_size_in = nonzero_size_in
        self.nonzero_size_out = nonzero_size_out
        super().__init__(name, default, optional, readonly)

    def coerce(self, instance, nengo_obj):  # pylint: disable=arguments-renamed
        nengo_objects = (
            NengoObject,
            ObjView,
            nengo.ensemble.Neurons,
            nengo.connection.LearningRule,
        )
        if not isinstance(nengo_obj, nengo_objects):
            raise ValidationError(
                f"'{nengo_obj}' is not a Nengo object", attr=self.name, obj=instance
            )
        if self.nonzero_size_in and nengo_obj.size_in < 1:
            raise ValidationError(
                f"'{nengo_obj}' must have size_in > 0.", attr=self.name, obj=instance
            )
        if self.nonzero_size_out and nengo_obj.size_out < 1:
            raise ValidationError(
                f"'{nengo_obj}' must have size_out > 0.", attr=self.name, obj=instance
            )
        return super().coerce(instance, nengo_obj)


class Process(FrozenObject):
    """A general system with input, output, and state.

    For more details on how to use processes and make
    custom process subclasses, see :doc:`examples/advanced/processes`.

    Parameters
    ----------
    default_size_in : int
        Sets the default size in for nodes using this process.
    default_size_out : int
        Sets the default size out for nodes running this process. Also,
        if ``d`` is not specified in `~.Process.run` or `~.Process.run_steps`,
        this will be used.
    default_dt : float
        If ``dt`` is not specified in `~.Process.run`, `~.Process.run_steps`,
        `~.Process.ntrange`, or `~.Process.trange`, this will be used.
    seed : int, optional
        Random number seed. Ensures random factors will be the same each run.

    Attributes
    ----------
    default_dt : float
        If ``dt`` is not specified in `~.Process.run`, `~.Process.run_steps`,
        `~.Process.ntrange`, or `~.Process.trange`, this will be used.
    default_size_in : int
        The default size in for nodes using this process.
    default_size_out : int
        The default size out for nodes running this process. Also, if ``d`` is
        not specified in `~.Process.run` or `~.Process.run_steps`,
        this will be used.
    seed : int or None
        Random number seed. Ensures random factors will be the same each run.
    """

    default_size_in = IntParam("default_size_in", low=0)
    default_size_out = IntParam("default_size_out", low=0)
    default_dt = NumberParam("default_dt", low=0, low_open=True)
    seed = IntParam("seed", low=0, high=maxseed, optional=True)

    def __init__(
        self, default_size_in=0, default_size_out=1, default_dt=0.001, seed=None
    ):
        super().__init__()
        self.default_size_in = default_size_in
        self.default_size_out = default_size_out
        self.default_dt = default_dt
        self.seed = seed

    def apply(self, x, d=None, dt=None, rng=np.random, copy=True, **kwargs):
        """Run process on a given input.

        Keyword arguments that do not appear in the parameter list below
        will be passed to the ``make_step`` function of this process.

        Parameters
        ----------
        x : ndarray
            The input signal given to the process.
        d : int, optional
            Output dimensionality. If None, ``default_size_out`` will be used.
        dt : float, optional
            Simulation timestep. If None, ``default_dt`` will be used.
        rng : `numpy.random.RandomState`
            Random number generator used for stochstic processes.
        copy : bool, optional
            If True, a new output array will be created for output.
            If False, the input signal ``x`` will be overwritten.
        """
        shape_in = as_shape(np.asarray(x[0]).shape, min_dim=1)
        shape_out = as_shape(self.default_size_out if d is None else d)
        dt = self.default_dt if dt is None else dt
        rng = self.get_rng(rng)
        state = self.make_state(shape_in, shape_out, dt)
        step = self.make_step(shape_in, shape_out, dt, rng, state, **kwargs)
        output = np.zeros((len(x),) + shape_out) if copy else x
        for i, xi in enumerate(x):
            output[i] = step((i + 1) * dt, xi)
        return output

    def get_rng(self, rng):
        """Get a properly seeded independent RNG for the process step.

        Parameters
        ----------
        rng : `numpy.random.RandomState`
            The parent random number generator to use if the seed is not set.
        """
        seed = rng.randint(maxint) if self.seed is None else self.seed
        return np.random.RandomState(seed)

    def make_state(self, shape_in, shape_out, dt, dtype=None):
        """Get a dictionary of signals to represent the state of this process.

        The builder uses this to allocate memory for the process state, so
        that the state can be represented as part of the whole simulator state.

        .. versionadded:: 3.0.0

        Parameters
        ----------
        shape_in : tuple
            The shape of the input signal.
        shape_out : tuple
            The shape of the output signal.
        dt : float
            The simulation timestep.
        dtype : `numpy.dtype`
            The data type requested by the builder. If `None`, then this
            function is free to choose the best type for the signals involved.

        Returns
        -------
        initial_state : {string: `numpy.ndarray`}
            A dictionary mapping keys to arrays containing the initial state
            values. The keys will be used to identify the signals in
            `.Process.make_step`.
        """
        return {}  # Implement if the process has a state

    def make_step(self, shape_in, shape_out, dt, rng, state):
        """Create function that advances the process forward one time step.

        This must be implemented by all custom processes. The parameters below
        indicate what information is provided by the builder.

        Parameters
        ----------
        shape_in : tuple
            The shape of the input signal.
        shape_out : tuple
            The shape of the output signal.
        dt : float
            The simulation timestep.
        rng : `numpy.random.RandomState`
            A random number generator.
        state : {string: `numpy.ndarray`}
            A dictionary mapping keys to signals, where the signals fully
            represent the state of the process. The signals are initialized
            by `.Process.make_state`.

            .. versionadded:: 3.0.0
        """
        raise NotImplementedError("Process must implement `make_step` method.")

    def run(self, t, d=None, dt=None, rng=np.random, **kwargs):
        """Run process without input for given length of time.

        Keyword arguments that do not appear in the parameter list below
        will be passed to the ``make_step`` function of this process.

        Parameters
        ----------
        t : float
            The length of time to run.
        d : int, optional
            Output dimensionality. If None, ``default_size_out`` will be used.
        dt : float, optional
            Simulation timestep. If None, ``default_dt`` will be used.
        rng : `numpy.random.RandomState`
            Random number generator used for stochstic processes.
        """
        dt = self.default_dt if dt is None else dt
        n_steps = int(np.round(float(t) / dt))
        return self.run_steps(n_steps, d=d, dt=dt, rng=rng, **kwargs)

    def run_steps(self, n_steps, d=None, dt=None, rng=np.random, **kwargs):
        """Run process without input for given number of steps.

        Keyword arguments that do not appear in the parameter list below
        will be passed to the ``make_step`` function of this process.

        Parameters
        ----------
        n_steps : int
            The number of steps to run.
        d : int, optional
            Output dimensionality. If None, ``default_size_out`` will be used.
        dt : float, optional
            Simulation timestep. If None, ``default_dt`` will be used.
        rng : `numpy.random.RandomState`
            Random number generator used for stochstic processes.
        """
        shape_in = as_shape(0)
        shape_out = as_shape(self.default_size_out if d is None else d)
        dt = self.default_dt if dt is None else dt
        rng = self.get_rng(rng)
        state = self.make_state(shape_in, shape_out, dt)
        step = self.make_step(shape_in, shape_out, dt, rng, state, **kwargs)
        output = np.zeros((n_steps,) + shape_out)
        for i in range(n_steps):
            output[i] = step((i + 1) * dt)
        return output

    def ntrange(self, n_steps, dt=None):
        """Create time points corresponding to a given number of steps.

        Parameters
        ----------
        n_steps : int
            The given number of steps.
        dt : float, optional
            Simulation timestep. If None, ``default_dt`` will be used.
        """
        dt = self.default_dt if dt is None else dt
        return dt * np.arange(1, n_steps + 1)

    def trange(self, t, dt=None):
        """Create time points corresponding to a given length of time.

        Parameters
        ----------
        t : float
            The given length of time.
        dt : float, optional
            Simulation timestep. If None, ``default_dt`` will be used.
        """
        dt = self.default_dt if dt is None else dt
        n_steps = int(np.round(float(t) / dt))
        return self.ntrange(n_steps, dt=dt)


class ProcessParam(Parameter):
    """Must be a Process."""

    def coerce(self, instance, process):  # pylint: disable=arguments-renamed
        self.check_type(instance, process, Process)
        return super().coerce(instance, process)



# Automatically generated by nengo-bones, do not edit this file directly

# pylint: disable=consider-using-f-string

"""Nengo version information.

We use semantic versioning (see http://semver.org/)
and conform to PEP440 (see https://www.python.org/dev/peps/pep-0440/).
'.dev0' will be added to the version unless the code base represents
a release version. Release versions are git tagged with the version.
"""

version_info = (3, 2, 0)

name = "nengo"
dev = None

# use old string formatting, so that this can still run in Python <= 3.5
# (since this file is parsed in setup.py, before python_requires is applied)
version = ".".join(str(v) for v in version_info)
if dev is not None:
    version += ".dev%d" % dev

copyright = "Copyright (c) 2013-2022 Applied Brain Research"



import warnings

import nengo
from nengo.spa.buffer import Buffer


class Memory(Buffer):
    """A SPA module capable of storing a vector over time.

    Parameters are the same as `.spa.Buffer`, with the addition of
    ``synapse`` and ``tau``.

    .. note:: Deprecated in Nengo 2.1.0. Use `.spa.State` instead.

    Parameters
    ----------
    dimensions : int
        Number of dimensions for the vector.
    subdimensions : int, optional
        Size of the individual ensembles making up the vector.
        Must divide ``dimensions`` evenly.
    neurons_per_dimensions : int, optional
        Number of neurons in an ensemble will be
        ``neurons_per_dimensions * subdimensions``.
    synapse : float, optional
        Synaptic filter to use on recurrent connection.
    vocab : Vocabulary, optional
        The vocabulary to use to interpret the vector. If None,
        the default vocabulary for the given dimensionality is used.
    tau : float or None, optional
        Effective time constant of the integrator. If None, it should
        have an infinite time constant.
    direct : bool, optional
        Whether or not to use direct mode for the neurons.

    label : str, optional
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional
        The seed used for random number generation.
    add_to_container : bool, optional
        Determines if this Network will be added to the current container.
        If None, will be true if currently within a Network.
    """

    def __init__(
        self,
        dimensions,
        subdimensions=16,
        neurons_per_dimension=50,
        synapse=0.01,
        vocab=None,
        tau=None,
        direct=False,
        label=None,
        seed=None,
        add_to_container=None,
    ):
        warnings.warn("Memory is deprecated in favour of spa.State", DeprecationWarning)
        super().__init__(
            dimensions=dimensions,
            subdimensions=subdimensions,
            neurons_per_dimension=neurons_per_dimension,
            vocab=vocab,
            direct=direct,
            label=label,
            seed=seed,
            add_to_container=add_to_container,
        )

        if tau is None:
            transform = 1.0
        else:
            transform = 1.0 - synapse / tau

        with self:
            nengo.Connection(
                self.state.output,
                self.state.input,
                transform=transform,
                synapse=synapse,
            )



import numpy as np
import pytest
from matplotlib import tri

import nengo
from nengo.utils.connection import eval_point_decoding
from nengo.utils.numpy import rms


@pytest.mark.parametrize("points_arg", [False, True])
def test_eval_point_decoding(points_arg, Simulator, NonDirectNeuronType, plt, seed):
    with nengo.Network(seed=seed) as model:
        model.config[nengo.Ensemble].neuron_type = NonDirectNeuronType()
        a = nengo.Ensemble(200, 2)
        b = nengo.Ensemble(100, 1)
        c = nengo.Connection(a, b, function=lambda x: x[0] * x[1])

    kwargs = {}
    if points_arg:
        x = np.linspace(-1, 1, 51)
        y = np.linspace(-1, 1, 51)
        X, Y = np.meshgrid(x, y)
        kwargs["eval_points"] = np.column_stack([X.ravel(), Y.ravel()])

    with Simulator(model) as sim:
        eval_points, targets, decoded = eval_point_decoding(c, sim, **kwargs)

    def contour(xy, z):
        xi, yi = np.meshgrid(np.linspace(-1, 1, 101), np.linspace(-1, 1, 101))
        triang = tri.Triangulation(xy[:, 0], xy[:, 1])
        interp_lin = tri.LinearTriInterpolator(triang, z.ravel())
        zi = interp_lin(xi, yi)
        plt.contourf(xi, yi, zi, cmap=plt.cm.seismic)
        plt.colorbar()

    plt.figure(figsize=(15, 5))
    plt.subplot(131)
    contour(eval_points, targets)
    plt.title("Target (desired decoding)")
    plt.subplot(132)
    plt.title("Actual decoding")
    contour(eval_points, decoded)
    plt.subplot(133)
    plt.title("Difference between actual and desired")
    contour(eval_points, decoded - targets)

    # Generous error check, just to make sure it's in the right ballpark.
    # Also make sure error is above zero, i.e. y != z
    error = rms(decoded - targets, axis=1).mean()
    assert 1e-8 < error < 0.1



import numpy as np
import pytest

import nengo
from nengo.networks.product import dot_product_transform
from nengo.utils.numpy import rms


def test_sine_waves(Simulator, plt, seed):
    radius = 2
    dim = 5
    product = nengo.networks.Product(200, dim, radius, seed=seed)

    func_a = lambda t: np.sqrt(radius) * np.sin(np.arange(1, dim + 1) * 2 * np.pi * t)
    func_b = lambda t: np.sqrt(radius) * np.sin(np.arange(dim, 0, -1) * 2 * np.pi * t)
    with product:
        input_a = nengo.Node(func_a)
        input_b = nengo.Node(func_b)
        nengo.Connection(input_a, product.input_a)
        nengo.Connection(input_b, product.input_b)
        p = nengo.Probe(product.output, synapse=0.005)

    with Simulator(product) as sim:
        sim.run(1.0)

    t = sim.trange()
    ideal = np.asarray([func_a(tt) for tt in t]) * np.asarray([func_b(tt) for tt in t])
    delay = 0.013
    offset = np.where(t >= delay)[0]

    for i in range(dim):
        plt.subplot(dim + 1, 1, i + 1)
        plt.plot(t + delay, ideal[:, i])
        plt.plot(t, sim.data[p][:, i])
        plt.xlim(right=t[-1])
        plt.yticks((-2, 0, 2))

    assert rms(ideal[: len(offset), :] - sim.data[p][offset, :]) < 0.2


def test_direct_mode_with_single_neuron(Simulator, plt, seed):
    radius = 2
    dim = 5

    config = nengo.Config(nengo.Ensemble)
    config[nengo.Ensemble].neuron_type = nengo.Direct()
    with config:
        product = nengo.networks.Product(1, dim, radius, seed=seed)

    func_a = lambda t: np.sqrt(radius) * np.sin(np.arange(1, dim + 1) * 2 * np.pi * t)
    func_b = lambda t: np.sqrt(radius) * np.sin(np.arange(dim, 0, -1) * 2 * np.pi * t)
    with product:
        input_a = nengo.Node(func_a)
        input_b = nengo.Node(func_b)
        nengo.Connection(input_a, product.input_a)
        nengo.Connection(input_b, product.input_b)
        p = nengo.Probe(product.output, synapse=0.005)

    with Simulator(product) as sim:
        sim.run(1.0)

    t = sim.trange()
    ideal = np.asarray([func_a(tt) for tt in t]) * np.asarray([func_b(tt) for tt in t])
    delay = 0.013
    offset = np.where(t >= delay)[0]

    for i in range(dim):
        plt.subplot(dim + 1, 1, i + 1)
        plt.plot(t + delay, ideal[:, i])
        plt.plot(t, sim.data[p][:, i])
        plt.xlim(right=t[-1])
        plt.yticks((-2, 0, 2))

    assert rms(ideal[: len(offset), :] - sim.data[p][offset, :]) < 0.2


@pytest.mark.parametrize("dims, scale", [(1, 3), (5, 2.4)])
def test_dot_product_transform(dims, scale):
    assert dot_product_transform(dims).shape == (1, dims)
    assert (dot_product_transform(dims, scale=scale) == scale).all()



"""Utilities to convert to and from bytes.

Used by nengo.rc in order to present file sizes to users in
human-readable formats.

This code adapted from
https://web.archive.org/web/20200817051754/http://code.activestate.com/recipes/578019-bytes-to-human-human-to-bytes-converter/?in=user-4178764
under the MIT License.
"""


def bytes2human(n, fmt="%(value).1f %(symbol)s"):
    """Convert from a size in bytes to a human readable string.

    Examples
    --------

    .. testcode::

       from nengo.utils.cache import bytes2human

       print(bytes2human(10000))
       print(bytes2human(100001221))

    .. testoutput::

       9.8 KB
       95.4 MB

    """
    symbols = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    prefix = {}
    for i, s in enumerate(symbols[1:]):
        prefix[s] = 1 << (i + 1) * 10
    for symbol in reversed(symbols[1:]):
        if n >= prefix[symbol]:
            value = float(n) / prefix[symbol]
            return fmt % {"value": value, "symbol": symbol}
    return fmt % {"value": n, "symbol": symbols[0]}


def human2bytes(s):
    """Convert from a human readable string to a size in bytes.

    Examples
    --------

    .. testcode::

       from nengo.utils.cache import human2bytes

       print(human2bytes('1 MB'))
       print(human2bytes('1 GB'))

    .. testoutput::

       1048576
       1073741824

    """
    symbols = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")

    ix = -1 if s[-2].isdigit() else -2
    letter = s[ix:].strip().upper()
    num = s[:ix].strip()
    assert letter in symbols
    num = float(num)
    prefix = {symbols[0]: 1}
    for i, symbol in enumerate(symbols[1:]):
        prefix[symbol] = 1 << (i + 1) * 10
    return int(num * prefix[letter])


def byte_align(size, alignment):
    """Returns the int larger than ``size`` aligned to ``alginment`` bytes."""
    mask = alignment - 1
    if size & mask == 0:
        return size
    else:
        return (size | mask) + 1



import pytest

from nengo import spa
from nengo.exceptions import SpaParseError
from nengo.spa.actions import Action, Actions, Effect, Expression


def test_expression():
    c = Expression(["a", "b"], "dot(a, A)")
    assert str(c.expression) == "dot(a, A)"

    c = Expression(["a", "b"], "0.5*(2*dot(a, A)-dot(b,B))-2")
    assert str(c.expression) == "dot(a, A) + -0.5 * dot(b, B) + -2"

    c = Expression(["a"], "1")
    assert str(c.expression) == "1"

    with pytest.raises(SpaParseError):
        c = Expression(["a", "b"], "dot(c, C)")


def test_scalars():
    c = Expression([], "1")
    assert str(c.expression) == "1"

    c = Expression([], "1 - 1")
    assert str(c.expression) == "0"

    c = Expression(["x"], "dot(1, x)")
    assert str(c.expression) == "dot(1, x)"

    c = Expression(["x"], "dot(x, -1) + 1")
    assert str(c.expression) == "dot(x, -1) + 1"

    c = Expression(["a", "b"], "2*dot(a, 1) - dot(b, -1) + dot(a, b)")
    assert str(c.expression) == "2 * dot(a, 1) + -dot(b, -1) + dot(a, b)"

    c = Expression(["a", "b"], "a*b - 1 + 2*b")
    assert str(c.expression) == "((a) * (b)) * 1 + -1 + 2 * b"


def test_effect():
    e = Effect(["a", "b"], ["m"], "m=A")
    assert str(e) == "m=A"
    e = Effect(["a", "b"], ["m", "n"], "m=A, n=B")
    assert str(e) == "m=A, n=B"
    e = Effect(["a", "b"], ["m", "n"], "m=a, n=b*2*A")
    assert str(e) == "m=a, n=(2 * A) * b"
    e = Effect(["a", "b"], ["m"], "m=0")
    assert str(e) == "m=0"

    # Check that multiple lvalue=rvalue parsing is working with commas
    e = Effect(["a", "b"], ["x", "y", "z"], "x=a,y=dot(a,b),z=b")
    assert str(e) == "x=a, y=dot(a, b), z=b"
    e = Effect(["a", "b"], ["foo", "bar"], "  foo = dot(a, b)  , bar = b")
    assert str(e) == "foo=dot(a, b), bar=b"

    with pytest.raises(SpaParseError):
        Effect(["a", "b"], ["q"], "q=z")

    with pytest.raises(SpaParseError):
        Effect(["a", "b"], ["q"], "q=a, q=b")  # lvalue appears twice


def test_inverted():
    with pytest.raises(SpaParseError):
        Effect(["b"], ["a"], "a = ~2*b")
    with pytest.raises(SpaParseError):
        Effect(["b"], ["a"], "a = ~2*C*b")


def test_action():
    a = Action(
        ["vision", "memory"],
        ["motor", "memory"],
        "dot(vision, DOG) --> motor=vision",
        "test_rule",
    )
    assert str(a.condition) == "dot(vision, DOG)"
    assert str(a.effect) == "motor=vision"

    a = Action(["vision", "memory"], ["motor", "memory"], "motor=vision*A", "test_rule")
    assert a.condition is None
    assert str(a.effect) == "motor=A * vision"

    with pytest.raises(SpaParseError):
        a = Action(
            ["vision", "memory"], ["motor", "memory"], "motor=vis*A", "test_action"
        )

    with pytest.raises(SpaParseError):
        a = Action(
            ["vision", "memory"],
            ["motor", "memory"],
            "0.5 --> motor=vis*A",
            "test_action",
        )

    with pytest.raises(SpaParseError):
        a = Action(
            ["vision", "memory"],
            ["motor", "memory"],
            "0.5*dot(mem, a) --> motor=B",
            name=None,
        )

    with pytest.raises(SpaParseError):
        a = Action(
            ["vision", "memory"],
            ["motor", "memory"],
            "0.5*dot(memory+1, vision) --> motor=B",
            name="test_action",
        )

    with pytest.raises(SpaParseError):
        a = Action(
            ["vision", "memory"], ["motor", "memory"], "motor2=B", name="test_action"
        )
    with pytest.raises(SpaParseError):
        a = Action(
            ["vision", "memory"],
            ["motor", "memory"],
            "motor=A, motor2=B",
            name="test_action",
        )


@pytest.mark.filterwarnings("ignore:The actions currently being added")
def test_actions():
    a = Actions(
        "dot(state, A) --> state=B",
        "dot(state, B) --> state=A",
        default="1.0 --> state=C",
    )
    assert a.count == 3

    model = spa.SPA()
    with model:
        model.state = spa.State(16)
    a.process(model)
    assert str(a.actions[0].condition) == "dot(state, A)"
    assert str(a.actions[0].effect) == "state=B"
    assert str(a.actions[1].condition) == "dot(state, B)"
    assert str(a.actions[1].effect) == "state=A"
    assert str(a.actions[2].condition) == "1.0"
    assert str(a.actions[2].effect) == "state=C"

    a.add("dot(state, D) --> state=E")
    a.add(added="dot(state, E) --> state=F")
    a.process(model)
    assert str(a.actions[2].condition) == "dot(state, D)"
    assert str(a.actions[2].effect) == "state=E"
    assert str(a.actions[3].condition) == "dot(state, E)"
    assert str(a.actions[3].effect) == "state=F"



from nengo import rc


def test_read_file(tmp_path):
    filepath = str(tmp_path / "test_rc_file.txt")

    # write the test file
    with open(filepath, "w", encoding="utf-8") as fh:
        fh.write("[test_rc_section]\nsetting0 = 3\nsetting1=5")

    del fh

    # read the test file using `rc.read_file`
    with open(filepath, "r", encoding="utf-8") as fh:
        rc.read_file(fh)

    assert int(rc["test_rc_section"]["setting0"]) == 3
    assert int(rc["test_rc_section"]["setting1"]) == 5



import logging
import time

import numpy as np
import pytest

import nengo
from nengo.exceptions import BuildError, SimulationError, ValidationError
from nengo.neurons import (
    LIF,
    AdaptiveLIF,
    AdaptiveLIFRate,
    Direct,
    Izhikevich,
    LIFRate,
    NeuronType,
    NeuronTypeParam,
    PoissonSpiking,
    RatesToSpikesNeuronType,
    RectifiedLinear,
    RegularSpiking,
    Sigmoid,
    SpikingRectifiedLinear,
    StochasticSpiking,
    Tanh,
)
from nengo.processes import WhiteSignal
from nengo.solvers import LstsqL2nz
from nengo.utils.ensemble import tuning_curves
from nengo.utils.matplotlib import implot, rasterplot
from nengo.utils.numpy import rms


# --- define a composite neuron type, used in `AnyNeuronType` (see `pytest_nengo.py`)
class SpikingTanh(RegularSpiking):
    def __init__(self, tau_ref=0.0025, amplitude=1.0, initial_state=None):
        super().__init__(
            base_type=Tanh(tau_ref=tau_ref),
            amplitude=amplitude,
            initial_state=initial_state,
        )

    @property
    def tau_ref(self):
        return self.base_type.tau_ref


def test_lif_builtin(rng, allclose):
    """Test that the dynamic model approximately matches the rates."""
    dt = 1e-3
    t_final = 1.0

    N = 10
    lif = LIF()
    gain, bias = lif.gain_bias(rng.uniform(80, 100, size=N), rng.uniform(-1, 1, size=N))

    x = np.arange(-2, 2, 0.1)
    J = gain * x[:, None] + bias

    voltage = np.zeros_like(J)
    reftime = np.zeros_like(J)

    spikes = np.zeros((int(t_final / dt),) + J.shape)
    for spikes_i in spikes:
        lif.step(dt, J, spikes_i, voltage, reftime)

    math_rates = lif.rates(x, gain, bias)
    sim_rates = spikes.mean(0)
    assert allclose(sim_rates, math_rates, atol=1, rtol=0.02)


@pytest.mark.parametrize("dt", (0.001, 0.002))
def test_lif(Simulator, plt, rng, dt, allclose):
    """Test that the dynamic model approximately matches the rates."""
    n = 5000
    x = 0.5
    encoders = np.ones((n, 1))
    max_rates = rng.uniform(low=10, high=200, size=n)
    intercepts = rng.uniform(low=-1, high=1, size=n)

    m = nengo.Network()
    with m:
        ins = nengo.Node(x)
        ens = nengo.Ensemble(
            n,
            dimensions=1,
            neuron_type=LIF(),
            encoders=encoders,
            max_rates=max_rates,
            intercepts=intercepts,
        )
        nengo.Connection(ins, ens.neurons, transform=np.ones((n, 1)), synapse=None)
        spike_probe = nengo.Probe(ens.neurons)
        voltage_probe = nengo.Probe(ens.neurons, "voltage")
        ref_probe = nengo.Probe(ens.neurons, "refractory_time")

    t_final = 1.0
    with Simulator(m, dt=dt) as sim:
        sim.run(t_final)

    i = 3
    plt.subplot(311)
    plt.plot(sim.trange(), sim.data[spike_probe][:, :i])
    plt.subplot(312)
    plt.plot(sim.trange(), sim.data[voltage_probe][:, :i])
    plt.subplot(313)
    plt.plot(sim.trange(), sim.data[ref_probe][:, :i])
    plt.ylim([-dt, ens.neuron_type.tau_ref + dt])

    # check rates against analytic rates
    math_rates = ens.neuron_type.rates(
        x, *ens.neuron_type.gain_bias(max_rates, intercepts)
    )
    spikes = sim.data[spike_probe]
    sim_rates = (spikes > 0).sum(0) / t_final
    logging.info("ME = %f", (sim_rates - math_rates).mean())
    logging.info("RMSE = %f", rms(sim_rates - math_rates) / (rms(math_rates) + 1e-20))
    assert np.sum(math_rates > 0) > 0.5 * n, "At least 50% of neurons must fire"
    assert allclose(sim_rates, math_rates, atol=1, rtol=0.02)

    # if voltage and ref time are non-constant, the probe is doing something
    assert np.abs(np.diff(sim.data[voltage_probe])).sum() > 1
    assert np.abs(np.diff(sim.data[ref_probe])).sum() > 1

    # compute spike counts after each timestep
    actual_counts = (spikes > 0).cumsum(axis=0)
    expected_counts = np.outer(sim.trange(), math_rates)
    assert (abs(actual_counts - expected_counts) < 1).all()


@pytest.mark.parametrize("min_voltage", [-np.inf, -1, 0])
def test_lif_min_voltage(Simulator, plt, min_voltage, seed):
    model = nengo.Network(seed=seed)
    with model:
        stim = nengo.Node(lambda t: np.sin(t * 4 * np.pi))
        ens = nengo.Ensemble(
            n_neurons=10, dimensions=1, neuron_type=LIF(min_voltage=min_voltage)
        )
        nengo.Connection(stim, ens, synapse=None)
        p_val = nengo.Probe(ens, synapse=0.01)
        p_voltage = nengo.Probe(ens.neurons, "voltage")

    with Simulator(model) as sim:
        sim.run(0.5)

    plt.subplot(2, 1, 1)
    plt.plot(sim.trange(), sim.data[p_val])
    plt.ylabel("Decoded value")
    plt.subplot(2, 1, 2)
    plt.plot(sim.trange(), sim.data[p_voltage])
    plt.ylabel("Voltage")

    if min_voltage < -10:
        assert np.min(sim.data[p_voltage]) < -10
    else:
        assert np.min(sim.data[p_voltage]) == min_voltage


def test_lif_zero_tau_ref(Simulator, allclose):
    """If we set tau_ref=0, we should be able to reach firing rate 1/dt."""
    dt = 1e-3
    max_rate = 1.0 / dt
    with nengo.Network() as m:
        ens = nengo.Ensemble(
            1, 1, encoders=[[1]], max_rates=[max_rate], neuron_type=LIF(tau_ref=0)
        )
        nengo.Connection(nengo.Node(output=10), ens)
        p = nengo.Probe(ens.neurons)
    with Simulator(m) as sim:
        sim.run(0.02)
    assert allclose(sim.data[p][1:], max_rate)


def test_alif_rate(Simulator, plt, allclose):
    n = 100
    max_rates = 50 * np.ones(n)
    intercepts = np.linspace(-0.99, 0.99, n)
    encoders = np.ones((n, 1))

    model = nengo.Network()
    with model:
        u = nengo.Node(output=0.5)
        a = nengo.Ensemble(
            n,
            1,
            max_rates=max_rates,
            intercepts=intercepts,
            encoders=encoders,
            neuron_type=AdaptiveLIFRate(),
        )
        nengo.Connection(u, a, synapse=None)
        ap = nengo.Probe(a.neurons)

    with Simulator(model) as sim:
        sim.run(2.0)

    t = sim.trange()
    rates = sim.data[ap]
    _, ref = tuning_curves(a, sim, inputs=0.5)

    ax = plt.subplot(211)
    implot(plt, t, intercepts[::-1], rates.T, ax=ax)
    ax.set_xlabel("time [s]")
    ax.set_ylabel("input")
    ax = plt.subplot(212)
    ax.plot(intercepts, ref[::-1].T, "k--")
    ax.plot(intercepts, rates[[1, 500, 1000, -1], ::-1].T)
    ax.set_xlabel("input")
    ax.set_xlabel("rate")

    # check that initial tuning curve is the same as LIF rates
    assert allclose(rates[1], ref, atol=0.1, rtol=1e-3)

    # check that curves in firing region are monotonically decreasing
    assert np.all(np.diff(rates[1:, intercepts < 0.4], axis=0) < 0)


def test_alif(Simulator, plt):
    """Test ALIF and ALIFRate by comparing them to each other"""

    n = 100
    max_rates = 50 * np.ones(n)
    intercepts = np.linspace(-0.99, 0.99, n)
    encoders = np.ones((n, 1))
    nparams = dict(tau_n=1, inc_n=10e-3)
    eparams = dict(
        n_neurons=n, max_rates=max_rates, intercepts=intercepts, encoders=encoders
    )

    model = nengo.Network()
    with model:
        u = nengo.Node(output=0.5)
        a = nengo.Ensemble(
            neuron_type=AdaptiveLIFRate(**nparams), dimensions=1, **eparams
        )
        b = nengo.Ensemble(neuron_type=AdaptiveLIF(**nparams), dimensions=1, **eparams)
        nengo.Connection(u, a, synapse=0)
        nengo.Connection(u, b, synapse=0)
        ap = nengo.Probe(a.neurons)
        bp = nengo.Probe(b.neurons)

    with Simulator(model) as sim:
        sim.run(2.0)

    t = sim.trange()
    a_rates = sim.data[ap]
    spikes = sim.data[bp]
    b_rates = nengo.Lowpass(0.04).filtfilt(spikes)

    tmask = (t > 0.1) & (t < 1.7)
    rel_rmse = rms(b_rates[tmask] - a_rates[tmask]) / rms(a_rates[tmask])

    ax = plt.subplot(311)
    implot(plt, t, intercepts[::-1], a_rates.T, ax=ax)
    ax.set_ylabel("input")
    ax = plt.subplot(312)
    implot(plt, t, intercepts[::-1], b_rates.T, ax=ax)
    ax.set_ylabel("input")
    ax = plt.subplot(313)
    implot(plt, t, intercepts[::-1], (b_rates - a_rates)[tmask].T, ax=ax)
    ax.set_xlabel("time [s]")
    ax.set_ylabel("input")

    assert rel_rmse < 0.07


def test_izhikevich(Simulator, plt, seed):
    """Smoke test for using Izhikevich neurons.

    Tests that the 6 parameter sets listed in the original paper can be
    simulated in Nengo (but doesn't test any properties of them).
    """
    with nengo.Network() as m:
        u = nengo.Node(output=WhiteSignal(0.6, high=10), size_out=1)

        # Seed the ensembles (not network) so we get the same sort of neurons
        ens_args = {"n_neurons": 4, "dimensions": 1, "seed": seed}
        rs = nengo.Ensemble(neuron_type=Izhikevich(), **ens_args)
        ib = nengo.Ensemble(
            neuron_type=Izhikevich(reset_voltage=-55, reset_recovery=4), **ens_args
        )
        ch = nengo.Ensemble(
            neuron_type=Izhikevich(reset_voltage=-50, reset_recovery=2), **ens_args
        )
        fs = nengo.Ensemble(neuron_type=Izhikevich(tau_recovery=0.1), **ens_args)
        lts = nengo.Ensemble(neuron_type=Izhikevich(coupling=0.25), **ens_args)
        rz = nengo.Ensemble(
            neuron_type=Izhikevich(tau_recovery=0.1, coupling=0.26), **ens_args
        )

        ensembles = (rs, ib, ch, fs, lts, rz)
        out = {}
        spikes = {}
        for ens in ensembles:
            nengo.Connection(u, ens)
            out[ens] = nengo.Probe(ens, synapse=0.05)
            spikes[ens] = nengo.Probe(ens.neurons)
        up = nengo.Probe(u)

    with Simulator(m, seed=seed + 1) as sim:
        sim.run(0.6)
    t = sim.trange()

    def plot(ens, title, ix):
        ax = plt.subplot(len(ensembles), 1, ix)
        plt.title(title)
        plt.plot(t, sim.data[out[ens]], c="k", lw=1.5)
        plt.plot(t, sim.data[up], c="k", ls=":")
        ax = ax.twinx()
        ax.set_yticks(())
        rasterplot(t, sim.data[spikes[ens]], ax=ax)

    plt.figure(figsize=(10, 12))
    plot(rs, "Regular spiking", 1)
    plot(ib, "Intrinsically bursting", 2)
    plot(ch, "Chattering", 3)
    plot(fs, "Fast spiking", 4)
    plot(lts, "Low-threshold spiking", 5)
    plot(rz, "Resonator", 6)


@pytest.mark.parametrize("max_rate,intercept", [(300.0, 0.0), (100.0, 1.1)])
def test_sigmoid_response_curves(Simulator, max_rate, intercept, allclose):
    """Check the sigmoid response curve monotonically increases.

    The sigmoid response curve should work fine:

    - if max rate > rate at inflection point and intercept < 1
    - if max rate < rate at inflection point and intercept > 1
    """
    with nengo.Network() as m:
        e = nengo.Ensemble(
            1, 1, neuron_type=Sigmoid(), max_rates=[max_rate], intercepts=[intercept]
        )

    with Simulator(m) as sim:
        pass
    _, y = nengo.utils.ensemble.response_curves(e, sim)
    assert allclose(np.max(y), max_rate)
    assert np.all(y > 0.0)
    assert np.all(np.diff(y) > 0.0)  # monotonically increasing


def test_tanh_invalid():
    """Ensures validation error is thrown when given a large max rate."""
    tanh = Tanh(tau_ref=0.5)
    with pytest.raises(ValidationError, match="Max rates must be below"):
        tanh.gain_bias(max_rates=np.array([100]), intercepts=np.array([0]))


@pytest.mark.filterwarnings("ignore:divide by zero")
def test_sigmoid_invalid(Simulator):
    """Check that invalid sigmoid ensembles raise an error."""

    sigmoid = Sigmoid(tau_ref=0.5)
    with pytest.raises(ValidationError, match="Max rates must be below"):
        sigmoid.gain_bias(max_rates=np.array([100]), intercepts=np.array([0]))

    for max_rate, intercept in [(300.0, 1.1), (300.0, 1.0), (100.0, 0.9), (100, 1.0)]:
        with nengo.Network() as net:
            nengo.Ensemble(
                1,
                1,
                neuron_type=Sigmoid(),
                max_rates=[max_rate],
                intercepts=[intercept],
            )
        with pytest.raises(BuildError, match="lead to neurons with negative"):
            with Simulator(net):
                pass


@pytest.mark.parametrize("base_type", [LIFRate(), RectifiedLinear(), Tanh()])
def test_spiking_types(Simulator, base_type, seed, plt, allclose):
    spiking_types = [RegularSpiking, PoissonSpiking, StochasticSpiking]
    atols = {
        RegularSpiking: 0.08,
        PoissonSpiking: 0.15,
        StochasticSpiking: 0.15,
    }

    n_neurons = 400
    delay = 5

    with nengo.Network(seed=seed) as net:
        u = nengo.Node(lambda t: np.sin(2 * np.pi * t))
        a = nengo.Ensemble(n_neurons, 1, neuron_type=base_type)
        nengo.Connection(u, a)
        u_p = nengo.Probe(u, synapse=0.001 * delay)
        a_p = nengo.Probe(a, synapse=0.001 * delay)

    with Simulator(net, seed=seed) as sim:
        sim.run(1.0)
    baseline = sim.data[a_p][delay:]

    results = {}
    for spiking_type in spiking_types:
        a.neuron_type = spiking_type(base_type)

        with Simulator(net, seed=seed + 1) as sim:
            timer = time.time()
            sim.run(1.0)
            timer = time.time() - timer

        results[spiking_type] = {
            "u": sim.data[u_p][:-delay],
            "x": sim.data[a_p][delay:],
        }
        plt.plot(
            sim.trange(), sim.data[a_p], label=f"{spiking_type.__name__}: t={timer:.3f}"
        )

    plt.title(base_type)
    plt.plot(sim.trange()[delay:], sim.data[u_p][:-delay], "k--")
    plt.legend(loc=3)

    for spiking_type in spiking_types:
        x = results[spiking_type]["x"]
        u = results[spiking_type]["u"]
        assert allclose(x, u, atol=atols[spiking_type]), spiking_type.__name__
        # check that spike noise is the target amount above the base spiking model noise
        assert allclose(x, baseline, atol=atols[spiking_type]), spiking_type.__name__


def test_dt_dependence(Simulator, NonDirectNeuronType, plt, seed, allclose):
    """Neurons should not wildly change with different dt."""
    freq = 10 * (2 * np.pi)
    input_signal = lambda t: [np.sin(freq * t), np.cos(freq * t)]

    with nengo.Network(seed=seed) as m:
        m.config[nengo.Ensemble].neuron_type = NonDirectNeuronType()
        u = nengo.Node(input_signal, size_out=2)
        pre = nengo.Ensemble(80, dimensions=2)
        square = nengo.Ensemble(80, dimensions=2)
        nengo.Connection(u, pre)
        nengo.Connection(pre, square, function=lambda x: x ** 2)

        activity_p = nengo.Probe(square.neurons, synapse=0.05, sample_every=0.001)
        out_p = nengo.Probe(square, synapse=0.05, sample_every=0.001)

    activity_data = []
    out_data = []
    dts = (0.0001, 0.001)
    colors = ("b", "g", "r")
    ax1 = plt.subplot(2, 1, 1)
    ax2 = plt.subplot(2, 1, 2)
    for c, dt in zip(colors, dts):
        with Simulator(m, dt=dt, seed=seed + 1) as sim:
            sim.run(0.1)
        t = sim.trange(sample_every=0.001)
        activity_data.append(sim.data[activity_p])
        out_data.append(sim.data[out_p])
        ax1.plot(t, sim.data[out_p], c=c)
        # Just plot 5 neurons
        ax2.plot(t, sim.data[activity_p][..., :5], c=c)

    ax1.set_xlim(right=t[-1])
    ax1.set_ylabel("Decoded output")
    ax2.set_xlim(right=t[-1])
    ax2.set_ylabel("Neural activity")

    assert rms(activity_data[0] - activity_data[1]) < ((1.0 / dts[-1]) * 0.01)
    assert allclose(out_data[0], out_data[1], atol=0.05)


@pytest.mark.parametrize(
    "Neuron", [LIF, LIFRate, RectifiedLinear, SpikingRectifiedLinear]
)
def test_amplitude(Simulator, seed, rng, plt, allclose, Neuron):
    amp = 0.1
    neuron0 = Neuron()
    neuron = Neuron(amplitude=amp)

    # check static
    x = np.linspace(-5, 30)
    y = amp * neuron0.rates(x, 1.0, 0.0)
    y2 = neuron.rates(x, 1.0, 0.0)
    assert allclose(y, y2, atol=1e-5)

    # check dynamic
    n = 100
    x = 1.0
    encoders = np.ones((n, 1))
    max_rates = rng.uniform(low=20, high=200, size=n)
    intercepts = rng.uniform(low=-1, high=0.8, size=n)

    with nengo.Network(seed=seed) as model:
        ins = nengo.Node(x)
        ens = nengo.Ensemble(
            n,
            dimensions=1,
            neuron_type=neuron,
            encoders=encoders,
            max_rates=max_rates,
            intercepts=intercepts,
        )
        nengo.Connection(ins, ens, synapse=None)
        spike_probe = nengo.Probe(ens.neurons)

    dt = 0.001
    t_final = 0.5
    with Simulator(model, dt=dt) as sim:
        sim.run(t_final)

    spikes = sim.data[spike_probe]
    r = spikes.sum(axis=0) * (dt / t_final)

    gain, bias = neuron0.gain_bias(max_rates, intercepts)
    r0 = amp * neuron0.rates(x, gain, bias).squeeze(axis=0)

    i = np.argsort(r0)
    plt.plot(r0[i])
    plt.plot(r[i])

    error = rms(r - r0) / rms(r0)
    assert (error < 0.02).all()


def test_reset(Simulator, NonDirectNeuronType, seed, allclose):
    """Make sure resetting actually resets."""
    m = nengo.Network(seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = NonDirectNeuronType()
        u = nengo.Node(WhiteSignal(0.3, high=10), size_out=2)
        ens = nengo.Ensemble(60, dimensions=2)
        square = nengo.Ensemble(60, dimensions=2)
        nengo.Connection(u, ens)
        nengo.Connection(
            ens, square, function=lambda x: x ** 2, solver=LstsqL2nz(weights=True)
        )
        square_p = nengo.Probe(square, synapse=0.01)

    with Simulator(m, seed=seed + 1) as sim:
        sim.run(0.1)
        sim.run(0.2)
        t1 = sim.trange()
        square1 = np.array(sim.data[square_p])

        sim.reset()
        sim.run(0.3)

    assert allclose(sim.trange(), t1)
    assert allclose(sim.data[square_p], square1)


def test_neurontypeparam():
    """NeuronTypeParam must be a neuron type."""

    class Test:
        ntp = NeuronTypeParam("ntp", default=None)

    inst = Test()
    inst.ntp = LIF()
    assert isinstance(inst.ntp, LIF)
    with pytest.raises(ValueError):
        inst.ntp = "a"


def test_frozen():
    """Test attributes inherited from FrozenObject"""
    a = LIF()
    b = LIF()
    c = LIF(tau_rc=0.3)
    d = Izhikevich()

    assert hash(a) == hash(a)
    assert hash(b) == hash(b)
    assert hash(c) == hash(c)
    assert hash(d) == hash(d)

    assert a == b
    assert hash(a) == hash(b)
    assert a != c
    assert hash(a) != hash(c)  # not guaranteed, but highly likely
    assert b != c
    assert hash(b) != hash(c)  # not guaranteed, but highly likely
    assert a != d
    assert hash(a) != hash(d)  # not guaranteed, but highly likely

    with pytest.raises(ValueError):
        a.tau_rc = 0.3
    with pytest.raises(ValueError):
        d.coupling = 8


@pytest.mark.filterwarnings("ignore:divide by zero")
def test_direct_mode_nonfinite_value(Simulator):
    with nengo.Network() as model:
        e1 = nengo.Ensemble(10, 1, neuron_type=Direct())
        e2 = nengo.Ensemble(10, 1)
        nengo.Connection(e1, e2, function=lambda x: 1.0 / x)

    with Simulator(model) as sim:
        with pytest.raises(SimulationError):
            sim.run(0.01)


def test_direct_max_rates_intercepts():
    assert Direct().max_rates_intercepts(None, None) == (None, None)


def test_direct_step_error():
    with pytest.raises(SimulationError, match="Direct mode neurons.*simulated"):
        Direct().step(None, None, None)


@pytest.mark.parametrize("generic", (True, False))
def test_gain_bias(rng, NonDirectNeuronType, generic, allclose):
    if NonDirectNeuronType == Sigmoid and generic:
        # the generic method doesn't work with sigmoid neurons (because they're
        # always positive). that's not a failure, because the sigmoid neurons
        # never need to use the generic method normally, so we'll just skip
        # it for this test.
        return

    n = 100
    max_rates = rng.uniform(300, 400, size=n)
    intercepts = rng.uniform(-0.5, 0.5, size=n)
    neuron_type = NonDirectNeuronType()
    tolerance = 0.1 if generic else 1e-8

    if generic:
        gain, bias = NeuronType.gain_bias(neuron_type, max_rates, intercepts)
    else:
        gain, bias = neuron_type.gain_bias(max_rates, intercepts)

    assert allclose(neuron_type.rates(1, gain, bias), max_rates, atol=tolerance)

    if NonDirectNeuronType == Sigmoid:
        threshold = 0.5 / neuron_type.tau_ref
    else:
        threshold = 0

    x = (intercepts - tolerance)[np.newaxis, :]
    assert np.all(neuron_type.rates(x, gain, bias) <= threshold)
    x = (intercepts + tolerance)[np.newaxis, :]
    assert np.all(neuron_type.rates(x, gain, bias) > threshold)

    if generic:
        max_rates0, intercepts0 = NeuronType.max_rates_intercepts(
            neuron_type, gain, bias
        )
    else:
        max_rates0, intercepts0 = neuron_type.max_rates_intercepts(gain, bias)

    assert allclose(max_rates, max_rates0, atol=tolerance)
    assert allclose(intercepts, intercepts0, atol=tolerance)


def test_invalid_generic_gain_bias():
    class TestNeuronType(NeuronType):  # pylint: disable=abstract-method
        rate = 0

        def rates(self, x, gain, bias):
            return np.ones((x.size, 1)) * self.rate

    neuron_type = TestNeuronType()
    with pytest.raises(ValidationError, match="Could not find max current"):
        neuron_type.gain_bias(max_rates=[1], intercepts=[0])

    neuron_type.rate = 1
    with pytest.raises(ValidationError, match="Could not find firing threshold"):
        neuron_type.gain_bias(max_rates=[1], intercepts=[0])


def test_current(rng, allclose):
    neuron_type = NeuronType()
    n_neurons = 20
    gain = rng.rand(n_neurons)
    bias = rng.rand(n_neurons)

    # 3 samples
    x = rng.rand(3)
    current = neuron_type.current(x, gain, bias)
    assert allclose(current, gain * x.reshape(-1, 1) + bias)
    assert current.shape == (3, n_neurons)

    # 10 samples, different values for each neuron
    x = rng.rand(10, n_neurons)
    current = neuron_type.current(x, gain, bias)
    assert allclose(current, gain * x + bias)
    assert current.shape == (10, n_neurons)

    with pytest.raises(ValidationError):
        # Incorrect second dimension
        x = rng.rand(10, 2)
        current = neuron_type.current(x, gain, bias)

    with pytest.raises(ValidationError):
        # Too many dimensions
        x = rng.rand(10, n_neurons, 1)
        current = neuron_type.current(x, gain, bias)


def test_rates_shaping(rng, NonDirectNeuronType):
    neuron_type = NonDirectNeuronType()
    n_neurons = 20
    gain = rng.rand(n_neurons)
    bias = rng.rand(n_neurons)

    # 3 samples
    x = rng.rand(3)
    rates = neuron_type.rates(x, gain, bias)
    assert rates.shape == (3, n_neurons)

    # 10 samples, different values for each neuron
    x = rng.rand(10, n_neurons)
    rates = neuron_type.rates(x, gain, bias)
    assert rates.shape == (10, n_neurons)

    with pytest.raises(ValidationError):
        # Incorrect second dimension
        x = rng.rand(10, 2)
        rates = neuron_type.rates(x, gain, bias)

    with pytest.raises(ValidationError):
        # Too many dimensions
        x = rng.rand(10, n_neurons, 1)
        rates = neuron_type.rates(x, gain, bias)


@pytest.mark.parametrize("neuron_type", [nengo.LIF, nengo.SpikingRectifiedLinear])
@pytest.mark.filterwarnings("ignore:divide by zero")
def test_initial_state(neuron_type, Simulator, seed, plt, allclose):
    n = 50
    initial_voltage = {
        "uniform": nengo.dists.Uniform(0, 1),
        "choice": nengo.dists.Choice([0.5]),
        "linspace": np.linspace(0.1, 0.9, n),
    }

    with nengo.Network(seed=seed) as net:
        probes = {}
        for key, voltage in initial_voltage.items():
            a = nengo.Ensemble(
                50,
                1,
                neuron_type=neuron_type(initial_state={"voltage": voltage}),
                gain=nengo.dists.Choice([1.0]),
                bias=nengo.dists.Choice([0.0]),
            )
            probes[key] = nengo.Probe(a.neurons, "voltage")

    with Simulator(net) as sim:
        sim.run(0.1)

    for i, (key, probe) in enumerate(probes.items()):
        plt.subplot(len(probes), 1, i + 1)
        plt.plot(sim.trange(), sim.data[probe][:, :15])

    for key, voltage in initial_voltage.items():
        probe = probes[key]
        initial_voltage = sim.data[probe][0]
        if neuron_type is nengo.LIF:  # undo decay
            initial_voltage = initial_voltage * np.exp(sim.dt / a.neuron_type.tau_rc)

        if isinstance(voltage, nengo.dists.Uniform):
            histogram, _ = np.histogram(initial_voltage, bins=10)
            histogram = histogram / histogram.mean()
            assert np.abs(histogram - 1).mean() < 0.5
        else:
            if isinstance(voltage, nengo.dists.Distribution):
                voltage = voltage.sample(n)
            assert allclose(initial_voltage, voltage, atol=1e-5)


def test_bad_initial_state(Simulator):
    with pytest.raises(ValidationError, match="State variable 'rates' not recognized"):
        nengo.LIF(initial_state={"rates": nengo.dists.Choice([0])})

    with pytest.raises(ValidationError, match="State variable 'voltage' must be a"):
        nengo.LIF(initial_state={"voltage": "voltage"})

    class MyLIF(nengo.LIF):
        def __init__(self, bad_state=False):
            super().__init__()
            self.bad_state = bad_state
            if bad_state:
                self.state = {"in": nengo.dists.Choice([0])}

        def make_state(self, n_neurons, rng=np.random, dtype=None):
            if self.bad_state:
                return super().make_state(n_neurons, rng=np.random, dtype=None)
            return {"rng": rng, "dict": {}}

    with nengo.Network() as net:
        nengo.Ensemble(1, 1, neuron_type=MyLIF(bad_state=True))

    with pytest.raises(BuildError, match="State name 'in' overlaps"):
        with Simulator(net):
            pass

    with nengo.Network() as net:
        nengo.Ensemble(1, 1, neuron_type=MyLIF())

    with pytest.raises(BuildError, match="State 'dict' is of type 'dict'. Only"):
        with Simulator(net):
            pass


def test_spikes_to_spikes_warning():
    with pytest.warns(UserWarning, match="type 'LIF', which is a spiking"):
        nengo.PoissonSpiking(nengo.LIF())


def test_ratestospikes_state_overlap():
    class CustomSpiking(RatesToSpikesNeuronType):  # pylint: disable=abstract-method
        state = {"adaptation": nengo.dists.Choice([0])}

    with pytest.raises(ValidationError, match="have an overlapping state variable"):
        CustomSpiking(nengo.AdaptiveLIFRate())


def test_step_math():
    class CustomNeuronType(NeuronType):
        def step(self, dt, J, output, **state):
            return dict(dt=dt, J=J, output=output, state=state)

    argnames = ["dt", "J", "output", "state1", "state2"]
    args = {argname: object() for argname in argnames}
    with pytest.warns(UserWarning, match="'step_math' has been renamed to 'step'"):
        result = CustomNeuronType().step_math(**args)

    for arg, val in args.items():
        assert (result["state"][arg] if arg.startswith("state") else result[arg]) is val


def test_probeable():
    def check_neuron_type(neuron_type, expected):
        assert neuron_type.probeable == expected
        ens = nengo.Ensemble(10, 1, neuron_type=neuron_type)
        assert ens.neurons.probeable == expected + ("input",)

    with nengo.Network():
        check_neuron_type(Direct(), ("output",))
        check_neuron_type(RectifiedLinear(), ("output",))
        check_neuron_type(SpikingRectifiedLinear(), ("output", "voltage"))
        check_neuron_type(Sigmoid(), ("output",))
        check_neuron_type(Tanh(), ("output",))
        check_neuron_type(LIFRate(), ("output",))
        check_neuron_type(LIF(), ("output", "voltage", "refractory_time"))
        check_neuron_type(AdaptiveLIFRate(), ("output", "adaptation"))
        check_neuron_type(
            AdaptiveLIF(), ("output", "voltage", "refractory_time", "adaptation")
        )
        check_neuron_type(Izhikevich(), ("output", "voltage", "recovery"))
        check_neuron_type(RegularSpiking(LIFRate()), ("output", "rate_out", "voltage"))
        check_neuron_type(
            StochasticSpiking(AdaptiveLIFRate()), ("output", "rate_out", "adaptation")
        )
        check_neuron_type(PoissonSpiking(LIFRate()), ("output", "rate_out"))



import numpy as np
import pytest

from nengo.builder.operator import Operator
from nengo.builder.signal import Signal
from nengo.utils.simulator import validate_ops


def test_validate_ops():
    """tests validate_ops, including may_share_memory"""

    base1 = Signal(initial_value=np.ones((10, 4)))
    base2 = Signal(initial_value=np.ones((10, 4)))
    view1_a = base1[:5, :]
    view1_b = base1[5:, :]
    view1_ab = base1[:6, :]

    ops = [Operator() for _ in range(3)]

    # non-overlapping sets is OK
    validate_ops(sets={base1: [ops[1]], base2: [ops[2]]}, ups=[], incs=[])

    # missing set is bad
    with pytest.raises(AssertionError):
        validate_ops(sets={base1: [ops[1]], base2: []}, ups=[], incs=[])

    # multiple sets is bad
    with pytest.raises(AssertionError):
        validate_ops(sets={base1: [ops[1]], base2: ops}, ups=[], incs=[])

    # set base and view is bad
    with pytest.raises(AssertionError):
        validate_ops(sets={base1: [ops[1]], view1_a: [ops[2]]}, ups=[], incs=[])

    # set non-overlapping views is OK
    validate_ops(sets={view1_a: [ops[1]], view1_b: [ops[2]]}, ups=[], incs=[])

    # set overlapping views is bad
    with pytest.raises(AssertionError):
        validate_ops(sets={view1_ab: [ops[1]], view1_b: [ops[2]]}, ups=[], incs=[])

    # non-overlapping updates is OK
    validate_ops(ups={base1: [ops[1]], base2: [ops[2]]}, sets=[], incs=[])

    # missing update is bad
    with pytest.raises(AssertionError):
        validate_ops(ups={base1: [ops[1]], base2: []}, sets=[], incs=[])

    # multiple updates is bad
    with pytest.raises(AssertionError):
        validate_ops(ups={base1: [ops[1]], base2: ops}, sets=[], incs=[])

    # update base and view is bad
    with pytest.raises(AssertionError):
        validate_ops(ups={base1: [ops[1]], view1_a: [ops[2]]}, sets=[], incs=[])

    # update non-overlapping views is OK
    validate_ops(ups={view1_a: [ops[1]], view1_b: [ops[2]]}, sets=[], incs=[])

    # update overlapping views is bad
    with pytest.raises(AssertionError):
        validate_ops(ups={view1_ab: [ops[1]], view1_b: [ops[2]]}, sets=[], incs=[])



# From https://github.com/scipy/scipy/blob/v1.3.0/scipy/special/_comb.pyx
def comb(N, k):
    """Compute N choose k"""
    if k > N or N < 0 or k < 0:
        return 0

    M = N + 1
    nterms = min(k, N - k)

    numerator = 1
    denominator = 1
    for j in range(1, nterms + 1):
        numerator *= M - j
        denominator *= j

    return numerator // denominator



import contextlib
import logging

import numpy as np

import nengo.utils.numpy as npext
from nengo.builder.builder import Builder
from nengo.connection import Connection
from nengo.ensemble import Ensemble
from nengo.network import Network
from nengo.node import Node
from nengo.probe import Probe
from nengo.utils.progress import Progress

logger = logging.getLogger(__name__)
nullcontext = contextlib.contextmanager(lambda: (yield))


@Builder.register(Network)  # noqa: C901
def build_network(model, network, progress=None):  # noqa: C901
    """Builds a `.Network` object into a model.

    The network builder does this by mapping each high-level object to its
    associated signals and operators one-by-one, in the following order:

    1. Ensembles, nodes, neurons
    2. Subnetworks (recursively)
    3. Connections, learning rules
    4. Probes

    Before calling any of the individual objects' build functions, random
    number seeds are assigned to objects that did not have a seed explicitly
    set by the user. Whether the seed was assigned manually or automatically
    is tracked, and the decoder cache is only used when the seed is assigned
    manually.

    Parameters
    ----------
    model : Model
        The model to build into.
    network : Network
        The network to build.
    progress : Progress, optional
        Object used to track the build progress.

        Note that this will only affect top-level networks.

    Notes
    -----
    Sets ``model.params[network]`` to ``None``.
    """
    if model.toplevel is None:
        model.toplevel = network
        seed_network(network, seeds=model.seeds, seeded=model.seeded)

        if progress is not None:
            # number of sub-objects, plus 1 to account for this network
            progress.max_steps = len(network.all_objects) + 1

            def build_callback(obj):
                if isinstance(obj, tuple(network.objects)):
                    progress.step()

            model.build_callback = build_callback

    if progress is None:
        progress = Progress()  # dummy progress

    # Set config
    old_config = model.config
    model.config = network.config

    # If this is the toplevel network, enter the decoder cache
    context = model.decoder_cache if model.toplevel is network else nullcontext()
    with context, progress:
        logger.debug("Network step 1: Building ensembles and nodes")
        for obj in network.ensembles + network.nodes:
            model.build(obj)

        logger.debug("Network step 2: Building subnetworks")
        for subnetwork in network.networks:
            model.build(subnetwork)

        logger.debug("Network step 3: Building connections")
        for conn in network.connections:
            # NB: we do these in the order in which they're defined, and build
            # the learning rule in the connection builder. Because learning
            # rules are attached to connections, the connection that contains
            # the learning rule (and the learning rule) are always built
            # *before* a connection that attaches to that learning rule.
            # Therefore, we don't have to worry about connection ordering here.
            # TODO: Except perhaps if the connection being learned
            # is in a subnetwork?
            model.build(conn)

        logger.debug("Network step 4: Building probes")
        for probe in network.probes:
            model.build(probe)

        if context is model.decoder_cache:
            model.decoder_cache.shrink()

        if model.toplevel is network:
            progress.step()
            model.build_callback = None

    # Unset config
    model.config = old_config
    model.params[network] = None


def seed_network(network, seeds, seeded, base_rng=np.random):
    """Populate seeding dictionaries for all objects in a network.

    This includes all subnetworks.

    .. versionadded:: 3.0.0

    Parameters
    ----------
    network : Network
        The network containing all objects to set seeds for.
    seeds : {object: int}
        Pre-existing map from objects to seeds for those objects.
        Will be modified in-place, but entries will not be
        overwritten if already set.
    seeded : {object: bool}
        Pre-existing map from objects to a boolean indicating whether they
        have a fixed seed either themselves or from a parent network (True),
        or whether the seed is randomly generated (False).
        Will be modified in-place, but entries will not be
        overwritten if already set.
    base_rng : np.random.RandomState
        Random number generator to use to set the seeds.
    """

    # seed this base network
    _set_seed(seeds, network, base_rng)
    _set_seeded(seeded, network)

    # seed all sub-objects
    _seed_network(network, seeds, seeded)


def _seed_network(network, seeds, seeded):
    """Recursive helper to set seeds for all child objects and subnetworks."""
    rng = np.random.RandomState(seeds[network])

    # Put probes last so that they don't influence other seeds
    sorted_types = (Connection, Ensemble, Network, Node, Probe)
    assert all(tp in sorted_types for tp in network.objects)

    # assign seeds to all child objects
    for obj_type in sorted_types:
        for obj in network.objects[obj_type]:
            _set_seed(seeds, obj, rng)
            _set_seeded(seeded, obj, parent=network)

    # assign seeds to subnetwork objects
    for subnetwork in network.networks:
        _seed_network(subnetwork, seeds, seeded)


def _set_seed(seeds, obj, rng):
    # Generate a seed no matter what, so that setting a seed or not on
    # one object doesn't affect the seeds of other objects.
    seed = rng.randint(npext.maxint)

    if obj in seeds:
        return  # do not overwrite an existing seed
    elif getattr(obj, "seed", None) is not None:
        seeds[obj] = obj.seed
    else:
        seeds[obj] = seed


def _set_seeded(seeded, obj, parent=None):
    # do not overwrite an existing value, since this value says how the
    # original seed was assigned (deterministically or randomly), and if we
    # re-determine this, we might be wrong (e.g. if obj.seed has changed)
    if obj not in seeded:
        seeded[obj] = (
            getattr(obj, "seed", None) is not None
            or parent is not None
            and seeded[parent]
        )



"""Operators represent calculations that will occur in the simulation.

This code adapted from sigops/operator.py and sigops/operators.py
(https://github.com/jaberg/sigops).
This modified code is included under the terms of their license:

Copyright (c) 2014, James Bergstra
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import numpy as np

import nengo.utils.numpy as npext
from nengo.exceptions import BuildError, SimulationError
from nengo.utils.functions import function_name


class Operator:
    """Base class for operator instances understood by Nengo.

    During one simulator timestep, a `.Signal` can experience

    1. at most one set operator (optional)
    2. any number of increments
    3. any number of reads
    4. at most one update

    in this specific order.

    A ``set`` defines the state of the signal at time :math:`t`, the start
    of the simulation timestep. That state can then be modified by
    ``increment`` operations. A signal's state will only be ``read`` after
    all increments are complete. The state is then finalized by an ``update``,
    which denotes the state that the signal should be at time :math:`t + dt`.

    Each operator must keep track of the signals that it manipulates,
    and which of these four types of manipulations is done to each signal
    so that the simulator can order all of the operators properly.

    .. note:: There are intentionally no valid default values for the
              `~.Operator.reads`, `~.Operator.sets`, `~.Operator.incs`,
              and `~.Operator.updates` properties to ensure that subclasses
              explicitly set these values.

    Parameters
    ----------
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    tag : str or None
        A label associated with the operator, for debugging purposes.
    """

    def __init__(self, tag=None):
        self.tag = tag

        self._sets = None
        self._incs = None
        self._reads = None
        self._updates = None

    def __repr__(self):
        tag_txt = "" if self.tag is None else f" '{self.tag}'"
        return f"<{type(self).__name__}{tag_txt} at 0x{id(self):x}>"

    def __str__(self):
        tag_txt = "" if self.tag is None else f" '{self.tag}'"
        return f"{type(self).__name__}{{{self._descstr}{tag_txt}}}"

    @property
    def _descstr(self):
        return ""

    @property
    def all_signals(self):
        return self.reads + self.sets + self.incs + self.updates

    @property
    def incs(self):
        """Signals incremented by this operator.

        Increments will be applied after sets (if it is set), and before reads.
        """
        return self._incs

    @incs.setter
    def incs(self, val):
        self._incs = val

    @property
    def reads(self):
        """Signals that are read and not modified by this operator.

        Reads occur after increments, and before updates.
        """
        return self._reads

    @reads.setter
    def reads(self, val):
        self._reads = val

    @property
    def sets(self):
        """Signals set by this operator.

        Sets occur first, before increments. A signal that is set here cannot
        be set or updated by any other operator.
        """
        return self._sets

    @sets.setter
    def sets(self, val):
        self._sets = val

    @property
    def updates(self):
        """Signals updated by this operator.

        Updates are the last operation to occur to a signal.
        """
        return self._updates

    @updates.setter
    def updates(self, val):
        self._updates = val

    def init_signals(self, signals):
        """Initialize the signals associated with this operator.

        The signals will be initialized into ``signals``.
        Operator subclasses that use extra buffers should create them here.

        Parameters
        ----------
        signals : SignalDict
            A mapping from signals to their associated live ndarrays.
        """
        for sig in self.all_signals:
            if sig not in signals:
                signals.init(sig)

    def make_step(self, signals, dt, rng):
        """Returns a callable that performs the desired computation.

        This method must be implemented by subclasses. To fully understand what
        an operator does, look at its implementation of ``make_step``.

        Parameters
        ----------
        signals : SignalDict
            A mapping from signals to their associated live ndarrays.
        dt : float
            Length of each simulation timestep, in seconds.
        rng : `numpy.random.RandomState`
            Random number generator for stochastic operators.
        """
        raise NotImplementedError("subclasses must implement this method.")


class TimeUpdate(Operator):
    """Updates the simulation step and time.

    Implements ``step[...] += 1`` and ``time[...] = step * dt``.

    A separate operator is used (rather than a combination of `.Copy` and
    `.DotInc`) so that other backends can manage these important parts of the
    simulation state separately from other signals.

    Parameters
    ----------
    step : Signal
        The signal associated with the integer step counter.
    time : Signal
        The signal associated with the time (a float, in seconds).
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    step : Signal
        The signal associated with the integer step counter.
    tag : str or None
        A label associated with the operator, for debugging purposes.
    time : Signal
        The signal associated with the time (a float, in seconds).

    Notes
    -----
    1. sets ``[step, time]``
    2. incs ``[]``
    3. reads ``[]``
    4. updates ``[]``
    """

    def __init__(self, step, time, tag=None):
        super().__init__(tag=tag)
        self.sets = [step, time]
        self.incs = []
        self.reads = []
        self.updates = []

    @property
    def step(self):
        return self.sets[0]

    @property
    def time(self):
        return self.sets[1]

    def make_step(self, signals, dt, rng):
        step = signals[self.step]
        time = signals[self.time]

        def step_timeupdate():
            step[...] += 1
            time[...] = step * dt

        return step_timeupdate


class Reset(Operator):
    """Assign a constant value to a Signal.

    Implements ``dst[...] = value``.

    Parameters
    ----------
    dst : Signal
        The Signal to reset.
    value : float, optional
        The constant value to which ``dst`` is set.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    dst : Signal
        The Signal to reset.
    tag : str or None
        A label associated with the operator, for debugging purposes.
    value : float
        The constant value to which ``dst`` is set.

    Notes
    -----
    1. sets ``[dst]``
    2. incs ``[]``
    3. reads ``[]``
    4. updates ``[]``
    """

    def __init__(self, dst, value=0, tag=None):
        super().__init__(tag=tag)
        self.value = float(value)

        self.sets = [dst]
        self.incs = []
        self.reads = []
        self.updates = []

    @property
    def dst(self):
        return self.sets[0]

    @property
    def _descstr(self):
        return str(self.dst)

    def make_step(self, signals, dt, rng):
        target = signals[self.dst]
        value = self.value

        def step_reset():
            target[...] = value

        return step_reset


class Copy(Operator):
    """Assign the value of one signal to another, with optional slicing.

    Implements:

    - ``dst[:] = src``
    - ``dst[dst_slice] = src[src_slice]``
      (when ``dst_slice`` or ``src_slice`` is not None)
    - ``dst[dst_slice] += src[src_slice]`` (when ``inc=True``)

    Parameters
    ----------
    dst : Signal
        The signal that will be assigned to (set).
    src : Signal
        The signal that will be copied (read).
    dst_slice : slice or list, optional
        Slice or list of indices associated with ``dst``.
    src_slice : slice or list, optional
        Slice or list of indices associated with ``src``
    inc : bool, optional
        Whether this should be an increment rather than a copy.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    dst : Signal
        The signal that will be assigned to (set).
    dst_slice : list or None
        Indices associated with ``dst``.
    src : Signal
        The signal that will be copied (read).
    src_slice : list or None
        Indices associated with ``src``.
    tag : str or None
        A label associated with the operator, for debugging purposes.

    Notes
    -----
    1. sets ``[] if inc else [dst]``
    2. incs ``[dst] if inc else []``
    3. reads ``[src]``
    4. updates ``[]``
    """

    def __init__(self, src, dst, src_slice=None, dst_slice=None, inc=False, tag=None):
        super().__init__(tag=tag)

        if isinstance(src_slice, slice):
            src = src[src_slice]
            src_slice = None
        if isinstance(dst_slice, slice):
            dst = dst[dst_slice]
            dst_slice = None
        # ^ src_slice and dst_slice are now either lists of indices or None

        self.src_slice = src_slice
        self.dst_slice = dst_slice
        self.inc = inc

        self.sets = [] if inc else [dst]
        self.incs = [dst] if inc else []
        self.reads = [src]
        self.updates = []

    @property
    def dst(self):
        return self.incs[0] if self.inc else self.sets[0]

    @property
    def src(self):
        return self.reads[0]

    @property
    def _descstr(self):
        def sigstring(sig, sl):
            return f"{sig}{f'[{sl}]' if sl is not None else ''}"

        return (
            f"{sigstring(self.src, self.src_slice)} -> "
            f"{sigstring(self.dst, self.dst_slice)}, inc={self.inc}"
        )

    def make_step(self, signals, dt, rng):
        src = signals[self.src]
        dst = signals[self.dst]
        src_slice = self.src_slice if self.src_slice is not None else Ellipsis
        dst_slice = self.dst_slice if self.dst_slice is not None else Ellipsis
        inc = self.inc

        # If there are repeated indices in dst_slice, special handling needed.
        repeats = False
        if npext.is_array_like(dst_slice):
            # copy because we might modify it
            dst_slice = np.array(dst_slice)
            if dst_slice.dtype.kind != "b":
                # get canonical, positive indices first
                dst_slice[dst_slice < 0] += len(dst)
                repeats = len(np.unique(dst_slice)) < len(dst_slice)

        if inc and repeats:

            def step_copy():
                np.add.at(dst, dst_slice, src[src_slice])

        elif inc:

            def step_copy():
                dst[dst_slice] += src[src_slice]

        elif repeats:
            raise BuildError(
                f"{self}: Cannot have repeated indices in "
                "``dst_slice`` when copy is not an increment"
            )
        else:

            def step_copy():
                dst[dst_slice] = src[src_slice]

        return step_copy


class ElementwiseInc(Operator):
    """Increment signal ``Y`` by ``A * X`` (with broadcasting).

    Implements ``Y[...] += A * X``.

    Parameters
    ----------
    A : Signal
        The first signal to be multiplied.
    X : Signal
        The second signal to be multiplied.
    Y : Signal
        The signal to be incremented.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    A : Signal
        The first signal to be multiplied.
    tag : str or None
        A label associated with the operator, for debugging purposes.
    X : Signal
        The second signal to be multiplied.
    Y : Signal
        The signal to be incremented.

    Notes
    -----
    1. sets ``[]``
    2. incs ``[Y]``
    3. reads ``[A, X]``
    4. updates ``[]``
    """

    def __init__(self, A, X, Y, tag=None):
        super().__init__(tag=tag)
        self.sets = []
        self.incs = [Y]
        self.reads = [A, X]
        self.updates = []

    @property
    def A(self):
        return self.reads[0]

    @property
    def X(self):
        return self.reads[1]

    @property
    def Y(self):
        return self.incs[0]

    @property
    def _descstr(self):
        return f"{self.A}, {self.X} -> {self.Y}"

    def make_step(self, signals, dt, rng):
        A = signals[self.A]
        X = signals[self.X]
        Y = signals[self.Y]

        # check broadcasting shapes
        Ashape = npext.broadcast_shape(A.shape, 2)
        Xshape = npext.broadcast_shape(X.shape, 2)
        Yshape = npext.broadcast_shape(Y.shape, 2)
        assert all(len(s) == 2 for s in [Ashape, Xshape, Yshape])
        for da, dx, dy in zip(Ashape, Xshape, Yshape):
            if not (da in [1, dy] and dx in [1, dy] and max(da, dx) == dy):
                raise BuildError(
                    "Incompatible shapes in ElementwiseInc: "
                    f"Trying to do {Yshape} += {Ashape} * {Xshape}"
                )

        def step_elementwiseinc():
            Y[...] += A * X

        return step_elementwiseinc


def reshape_dot(A, X, Y, tag=None):
    """Checks if the dot product needs to be reshaped.

    Also does a bunch of error checking based on the shapes of A and X.
    """
    badshape = False
    ashape = (1,) if A.shape == () else A.shape
    xshape = (1,) if X.shape == () else X.shape

    if A.shape == ():
        incshape = X.shape
    elif X.shape == ():
        incshape = A.shape
    elif X.ndim == 1:
        badshape = ashape[-1] != xshape[0]
        incshape = ashape[:-1]
    else:
        badshape = ashape[-1] != xshape[-2]
        incshape = ashape[:-1] + xshape[:-2] + xshape[-1:]

    if (badshape or incshape != Y.shape) and incshape != ():
        raise BuildError(f"shape mismatch in {tag}: {A.shape} x {X.shape} -> {Y.shape}")

    # Reshape to handle case when np.dot(A, X) and Y are both scalars
    return A.dot(X).size == Y.size == 1


class DotInc(Operator):
    """Increment signal ``Y`` by ``dot(A, X)``.

    Implements ``Y[...] += np.dot(A, X)``.

    .. note:: Currently, this only supports matrix-vector multiplies
              for compatibility with NengoOCL.

    Parameters
    ----------
    A : Signal
        The first signal to be multiplied (a matrix).
    X : Signal
        The second signal to be multiplied (a vector).
    Y : Signal
        The signal to be incremented.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    A : Signal
        The first signal to be multiplied.
    tag : str or None
        A label associated with the operator, for debugging purposes.
    X : Signal
        The second signal to be multiplied.
    Y : Signal
        The signal to be incremented.

    Notes
    -----
    1. sets ``[]``
    2. incs ``[Y]``
    3. reads ``[A, X]``
    4. updates ``[]``
    """

    def __init__(self, A, X, Y, reshape=None, tag=None):
        super().__init__(tag=tag)

        if X.ndim >= 2 and any(d > 1 for d in X.shape[1:]):
            raise BuildError("X must be a column vector")
        if Y.ndim >= 2 and any(d > 1 for d in Y.shape[1:]):
            raise BuildError("Y must be a column vector")

        self.reshape = reshape
        if self.reshape is None:
            self.reshape = reshape_dot(
                A.initial_value, X.initial_value, Y.initial_value, self.tag
            )

        self.sets = []
        self.incs = [Y]
        self.reads = [A, X]
        self.updates = []

    @property
    def A(self):
        return self.reads[0]

    @property
    def X(self):
        return self.reads[1]

    @property
    def Y(self):
        return self.incs[0]

    @property
    def _descstr(self):
        return f"{self.A}, {self.X} -> {self.Y}"

    def make_step(self, signals, dt, rng):
        X = signals[self.X]
        A = signals[self.A]
        Y = signals[self.Y]

        if self.reshape:

            def step_dotinc_reshape():
                Y[...] += A.dot(X).reshape(Y.shape)

            return step_dotinc_reshape

        else:

            def step_dotinc():
                Y[...] += A.dot(X)

            return step_dotinc


class SparseDotInc(DotInc):
    """Like `.DotInc` but ``A`` is a sparse matrix.

    .. versionadded:: 3.0.0
    """

    def __init__(self, A, X, Y, tag=None):
        # Disallow reshaping
        super().__init__(A, X, Y, reshape=False, tag=tag)

        if not A.sparse:
            raise BuildError(f"{self}: A must be a sparse Signal")


class BsrDotInc(DotInc):
    """Increment signal Y by dot(A, X) using block sparse row format.

    Implements ``Y[...] += np.dot(A, X)``, where ``A`` is an instance
    of `scipy.sparse.bsr_matrix`.

    .. note:: Requires SciPy.

    .. note:: Currently, this only supports matrix-vector multiplies
              for compatibility with NengoOCL.

    Parameters
    ----------
    A : (k, r, c) Signal
        The signal providing the k data blocks with r rows and c columns.
    X : (k * c) Signal
        The signal providing the k column vectors to multiply with.
    Y : (k * r) Signal
        The signal providing the k column vectors to update.
    indices : ndarray
        Column indices, see `scipy.sparse.bsr_matrix` for details.
    indptr : ndarray
        Column index pointers, see `scipy.sparse.bsr_matrix` for details.
    reshape : bool
        Whether to reshape the result.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    A : (k, r, c) Signal
        The signal providing the k data blocks with r rows and c columns.
    indices : ndarray
        Column indices, see `scipy.sparse.bsr_matrix` for details.
    indptr : ndarray
        Column index pointers, see `scipy.sparse.bsr_matrix` for details.
    reshape : bool
        Whether to reshape the result.
    tag : str or None
        A label associated with the operator, for debugging purposes.
    X : (k * c) Signal
        The signal providing the k column vectors to multiply with.
    Y : (k * r) Signal
        The signal providing the k column vectors to update.

    Notes
    -----
    1. sets ``[]``
    2. incs ``[Y]``
    3. reads ``[A, X]``
    4. updates ``[]``
    """

    def __init__(self, A, X, Y, indices, indptr, reshape=None, tag=None):
        from scipy.sparse import bsr_matrix  # pylint: disable=import-outside-toplevel

        self.bsr_matrix = bsr_matrix

        super().__init__(A, X, Y, reshape=reshape, tag=tag)

        self.indices = indices
        self.indptr = indptr

    def make_step(self, signals, dt, rng):
        X = signals[self.X]
        A = signals[self.A]
        Y = signals[self.Y]
        mat_A = self.bsr_matrix((A, self.indices, self.indptr))

        def step_dotinc():
            inc = mat_A.dot(X)
            if self.reshape:
                inc = inc.reshape(Y.shape)
            Y[...] += inc

        return step_dotinc


class SimPyFunc(Operator):
    """Apply a Python function to a signal, with optional arguments.

    Implements ``output[...] = fn(*args)`` where ``args`` can
    include the current simulation time ``t`` and an input signal ``x``.

    Note that ``output`` may also be None, in which case the function is
    called but no output is captured.

    Parameters
    ----------
    output : Signal or None
        The signal to be set. If None, the function is still called.
    fn : callable
        The function to call.
    t : Signal or None
        The signal associated with the time (a float, in seconds).
        If None, the time will not be passed to ``fn``.
    x : Signal or None
        An input signal to pass to ``fn``.
        If None, an input signal will not be passed to ``fn``.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    fn : callable
        The function to call.
    output : Signal or None
        The signal to be set. If None, the function is still called.
    t : Signal or None
        The signal associated with the time (a float, in seconds).
        If None, the time will not be passed to ``fn``.
    tag : str or None
        A label associated with the operator, for debugging purposes.
    x : Signal or None
        An input signal to pass to ``fn``.
        If None, an input signal will not be passed to ``fn``.

    Notes
    -----
    1. sets ``[] if output is None else [output]``
    2. incs ``[]``
    3. reads ``([] if t is None else [t]) + ([] if x is None else [x])``
    4. updates ``[]``
    """

    def __init__(self, output, fn, t, x, tag=None):
        super().__init__(tag=tag)
        self.fn = fn
        self.t_passed = t is not None
        self.x_passed = x is not None

        self.sets = [] if output is None else [output]
        self.incs = []
        self.reads = ([] if t is None else [t]) + ([] if x is None else [x])
        self.updates = []

    @property
    def output(self):
        if len(self.sets) == 1:
            return self.sets[0]
        return None

    @property
    def t(self):
        return self.reads[0] if self.t_passed else None

    @property
    def x(self):
        return self.reads[-1] if self.x_passed else None

    @property
    def _descstr(self):
        return f"{self.x} -> {self.output}, fn='{function_name(self.fn)}'"

    def make_step(self, signals, dt, rng):
        fn = self.fn
        output = signals[self.output] if self.output is not None else None
        t = signals[self.t] if self.t is not None else None
        x = signals[self.x] if self.x is not None else None

        def step_simpyfunc():
            args = (np.copy(x),) if x is not None else ()
            y = fn(t.item(), *args) if t is not None else fn(*args)

            if output is not None:
                try:
                    # required since Numpy turns None into NaN
                    if y is None or not np.all(np.isfinite(y)):
                        raise SimulationError(
                            f"Function '{function_name(self.fn)}' returned "
                            "non-finite value"
                        )

                    output[...] = y

                except (TypeError, ValueError) as e:
                    raise SimulationError(
                        f"Function '{function_name(self.fn)}' returned a value "
                        f"{y!r} of invalid type {type(y)!r}"
                    ) from e

        return step_simpyfunc



"""Utilities for progress tracking and display to the user."""

import importlib
import os
import sys
import threading
import time
import uuid
import warnings
from datetime import timedelta
from html import escape
from shutil import get_terminal_size

import numpy as np

from ..exceptions import ValidationError
from ..rc import rc
from .ipython import check_ipy_version, get_ipython

if get_ipython() is not None:
    from IPython.display import Javascript, display  # pragma: no cover


class MemoryLeakWarning(UserWarning):
    pass


warnings.filterwarnings("once", category=MemoryLeakWarning)


def timestamp2timedelta(timestamp):
    """Convert timestamp to timedelta."""

    if timestamp == -1:
        return "Unknown"
    return timedelta(seconds=np.ceil(timestamp))


def _load_class(name):
    mod_name, cls_name = name.rsplit(".", 1)
    mod = importlib.import_module(mod_name)
    return getattr(mod, cls_name)


class Progress:
    """Stores and tracks information about the progress of some process.

    This class is to be used as part of a ``with`` statement. Use ``step()`` to
    update the progress.

    Parameters
    ----------
    max_steps : int
        The total number of calculation steps of the process.
    name_during : str, optional
        Short description of the task to be used while it is running.
    name_after : str, optional
        Short description of the task to be used after it has
        finished. Defaults to ``name_during``.

    Attributes
    ----------
    max_steps : int, optional
        The total number of calculation steps of the process, if known.
    name_after : str
        Name of the task to be used after it has finished.
    name_during : str
        Name of the task to be used while it is running.
    steps : int
        Number of completed steps.
    success : bool or None
        Whether the process finished successfully. ``None`` if the process
        did not finish yet.
    time_end : float
        Time stamp of the time the process was finished or aborted.
    time_start : float
        Time stamp of the time the process was started.

    Examples
    --------

    .. testcode::

       from nengo.utils.progress import Progress

       max_steps = 10
       with Progress(max_steps=max_steps) as progress:
           for i in range(max_steps):
               # do something
               progress.step()

    """

    def __init__(self, name_during="", name_after=None, max_steps=None):
        if max_steps is not None and max_steps <= 0:
            raise ValidationError(
                f"must be at least 1 (got {max_steps})", attr="max_steps"
            )
        self.n_steps = 0
        self.max_steps = max_steps
        self.name_during = name_during
        if name_after is None:
            name_after = name_during
        self.name_after = name_after
        self.time_start = self.time_end = time.time()
        self.finished = False
        self.success = None

    @property
    def progress(self):
        """The current progress as a number from 0 to 1 (inclusive).

        Returns
        -------
        float
        """
        if self.max_steps is None:
            return 0.0
        return min(1.0, self.n_steps / self.max_steps)

    def elapsed_seconds(self):
        """The number of seconds passed since entering the ``with`` statement.

        Returns
        -------
        float
        """
        if self.finished:
            return self.time_end - self.time_start
        else:
            return time.time() - self.time_start

    def eta(self):
        """The estimated number of seconds until the process is finished.

        Stands for estimated time of arrival (ETA).
        If no estimate is available -1 will be returned.

        Returns
        -------
        float
        """
        if self.progress > 0.0:
            return (1.0 - self.progress) * self.elapsed_seconds() / self.progress
        else:
            return -1

    def __enter__(self):
        self.finished = False
        self.success = None
        self.n_steps = 0
        self.time_start = time.time()
        return self

    def __exit__(self, exc_type, dummy_exc_value, dummy_traceback):
        self.success = exc_type is None
        if self.success and self.max_steps is not None:
            self.n_steps = self.max_steps
        self.time_end = time.time()
        self.finished = True

    def step(self, n=1):
        """Advances the progress.

        Parameters
        ----------
        n : int
            Number of steps to advance the progress by.
        """
        self.n_steps += n


class ProgressBar:
    """Visualizes the progress of a process.

    This is an abstract base class that progress bar classes some inherit from.
    Progress bars should visually displaying the progress in some way.
    """

    def update(self, progress):
        """Updates the displayed progress.

        Parameters
        ----------
        progress : Progress
            The progress information to display.
        """
        raise NotImplementedError()

    def close(self):
        """Closes the progress bar.

        Indicates that not further updates will be made.
        """


class NoProgressBar(ProgressBar):
    """A progress bar that does not display anything.

    Helpful in headless situations or when using Nengo as a library.
    """

    def update(self, progress):
        pass


class TerminalProgressBar(ProgressBar):
    """A progress bar that is displayed as ASCII output on ``stdout``."""

    def update(self, progress):
        if progress.finished:
            line = self._get_finished_line(progress)
        elif progress.max_steps is None:
            line = self._get_unknown_progress_line(progress)
        else:
            line = self._get_in_progress_line(progress)
        sys.stdout.write(line)
        sys.stdout.flush()

    def _get_in_progress_line(self, progress):
        line = f"[{{}}] ETA: {timestamp2timedelta(progress.eta())}"
        percent_str = f" {progress.name_during}... {int(100 * progress.progress)}% "
        width, _ = get_terminal_size()
        progress_width = max(0, width - len(line))
        progress_str = (int(progress_width * progress.progress) * "#").ljust(
            progress_width
        )

        percent_pos = (len(progress_str) - len(percent_str)) // 2
        if percent_pos > 0:
            progress_str = (
                progress_str[:percent_pos]
                + percent_str
                + progress_str[percent_pos + len(percent_str) :]
            )

        return "\r" + line.format(progress_str)

    def _get_unknown_progress_line(self, progress):
        """Generates a progress line with continuously moving marker.

        This is to indicate processing while not knowing how far along we
        progressed with the processing.
        """
        duration = progress.elapsed_seconds()
        line = f"[{{}}] duration: {timestamp2timedelta(duration)}"
        text = f" {progress.name_during}... "
        width, _ = get_terminal_size()
        marker = ">>>>"
        progress_width = max(0, width - len(line) + 2)
        index_width = progress_width + len(marker)
        i = int(10.0 * duration) % (index_width + 1)
        progress_str = (" " * i) + marker + (" " * (index_width - i))
        progress_str = progress_str[len(marker) : -len(marker)]
        text_pos = (len(progress_str) - len(text)) // 2
        progress_str = (
            progress_str[:text_pos] + text + progress_str[text_pos + len(text) :]
        )
        return "\r" + line.format(progress_str)

    def _get_finished_line(self, progress):
        width, _ = get_terminal_size()
        elapsed_seconds = timestamp2timedelta(progress.elapsed_seconds())
        line = f"{progress.name_after} finished in {elapsed_seconds}.".ljust(width)
        return "\r" + line

    def close(self):
        sys.stdout.write(os.linesep)
        sys.stdout.flush()


class VdomProgressBar(ProgressBar):  # pragma: no cover
    """A progress bar using a virtual DOM representation.

    This HTML representation can be used in Jupyter lab (>=0.32) environments.

    .. versionadded: 3.0.0
    """

    def __init__(self):
        super().__init__()
        self._uuid = uuid.uuid4()
        self._handle = None
        self.progress = None

    def update(self, progress):
        self.progress = progress
        if self._handle is None:
            self._handle = display(self, display_id=True)
        else:
            self._handle.update(self)

    def _repr_mimebundle_(self, include, exclude, **kwargs):
        return {"application/vdom.v1+json": self._get_vdom(self.progress)}

    def _get_vdom(self, progress):
        return {
            "tagName": "div",
            "attributes": {
                "id": str(self._uuid),
                "style": {
                    "width": "100%",
                    "boxSizing": "border-box",
                    "border": "1px solid #cfcfcf",
                    "borderRadius": "4px",
                    "textAlign": "center",
                    "position": "relative",
                },
            },
            "children": [
                {
                    "tagName": "div",
                    "attributes": {
                        "class": "pb-text",
                        "style": {"position": "absolute", "width": "100%"},
                    },
                    "children": [self._get_text(self.progress)],
                },
                {
                    "tagName": "div",
                    "attributes": {
                        "class": "pb-fill",
                        "style": self._get_fill_style(self.progress),
                    },
                    "children": [
                        {
                            "tagName": "style",
                            "attributes": {"type": "text/css", "scoped": "scoped"},
                            "children": [
                                """
                        @keyframes pb-fill-anim {
                            0% { background-position: 0 0; }
                            100% { background-position: 100px 0; }
                        }}"""
                            ],
                        },
                        "\u00A0",  # non-breaking space
                    ],
                },
            ],
        }

    def _get_text(self, progress):
        if progress is None:
            text = ""
        elif progress.finished:
            text = (
                "{} finished in {}.".format(  # pylint: disable=consider-using-f-string
                    escape(progress.name_after),
                    timestamp2timedelta(progress.elapsed_seconds()),
                )
            )
        elif progress.max_steps is None:
            text = "{task}\u2026 duration: {duration}".format(  # pylint: disable=consider-using-f-string
                task=escape(progress.name_during),
                duration=timestamp2timedelta(progress.elapsed_seconds()),
            )
        else:
            text = "{task}\u2026 {progress:.0f}%, ETA: {eta}".format(  # pylint: disable=consider-using-f-string
                task=escape(progress.name_during),
                progress=100.0 * progress.progress,
                eta=timestamp2timedelta(progress.eta()),
            )
        return text

    def _get_fill_style(self, progress):
        if progress.max_steps is None:
            style = self._get_unknown_steps_fill_style(progress)
        else:
            style = self._get_known_steps_fill_style(progress)

        if progress.finished:
            style["animation"] = "none"
            style["backgroundImage"] = "none"

        return style

    def _get_known_steps_fill_style(self, progress):
        return {
            "width": f"{100.0 * progress.progress:.0f}%",
            "animation": "none",
            "backgroundColor": "#bdd2e6",
            "backgroundImage": "none",
            "transition": "width 0.1s linear" if progress.progress > 0.0 else "none",
        }

    def _get_unknown_steps_fill_style(self, progress):
        return {
            "width": "100%",
            "animation": "pb-fill-anim 2s linear infinite",
            "backgroundColor": "#bdd2e6",
            "backgroundSize": "100px 100%",
            "backgroundImage": (
                "repeating-linear-gradient("
                "90deg, #bdd2e6, #edf2f8 40%, #bdd2e6 80%, #bdd2e6)"
            ),
        }


class HtmlProgressBar(ProgressBar):  # pragma: no cover
    """A progress bar using a HTML representation.

    This HTML representation can be used in Jupyter notebook environments
    and is provided by the *_repr_html_* method that will be automatically
    used by IPython interpreters.

    If the kernel frontend does not support HTML (e.g., in Jupyter qtconsole),
    a warning message will be issued as the ASCII representation.
    """

    def __init__(self):
        super().__init__()
        self._uuid = uuid.uuid4()
        self._handle = None

    def update(self, progress):
        if self._handle is None:
            display(self._HtmlBase(self._uuid))
            self._handle = display(self._js_update(progress), display_id=True)
        else:
            self._handle.update(self._js_update(progress))

    class _HtmlBase:
        def __init__(self, my_uuid):
            self.uuid = my_uuid

        def __repr__(self):
            return (
                "HtmlProgressBar cannot be displayed. Please use the "
                "TerminalProgressBar. It can be enabled with `nengo.rc['progress']"
                "['progress_bar'] = 'nengo.utils.progress.TerminalProgressBar'`."
            )

        def _repr_html_(self):
            return """
                <script>
                    if (Jupyter.version.split(".")[0] < 5) {{
                        var pb = document.getElementById("{uuid}");
                        var text = document.createTextNode(
                            "HMTL progress bar requires Jupyter Notebook >= " +
                            "5.0 or Jupyter Lab. Alternatively, you can use " +
                            "TerminalProgressBar().");
                        pb.parentNode.insertBefore(text, pb);
                    }}
                </script>
                <div id="{uuid}" style="
                    width: 100%;
                    border: 1px solid #cfcfcf;
                    border-radius: 4px;
                    text-align: center;
                    position: relative;">
                  <div class="pb-text" style="
                      position: absolute;
                      width: 100%;">
                    0%
                  </div>
                  <div class="pb-fill" style="
                      background-color: #bdd2e6;
                      width: 0%;">
                    <style type="text/css" scoped="scoped">
                        @keyframes pb-fill-anim {{
                            0% {{ background-position: 0 0; }}
                            100% {{ background-position: 100px 0; }}
                        }}
                    </style>
                    &nbsp;
                  </div>
                </div>""".format(
                uuid=self.uuid
            )

    def _js_update(self, progress):
        if progress is None:
            text = ""
        elif progress.finished:
            text = (
                "{} finished in {}.".format(  # pylint: disable=consider-using-f-string
                    escape(progress.name_after),
                    timestamp2timedelta(progress.elapsed_seconds()),
                )
            )
        elif progress.max_steps is None:
            text = "{task}&hellip; duration: {duration}".format(  # pylint: disable=consider-using-f-string
                task=escape(progress.name_during),
                duration=timestamp2timedelta(progress.elapsed_seconds()),
            )
        else:
            text = "{task}&hellip; {progress:.0f}%, ETA: {eta}".format(  # pylint: disable=consider-using-f-string
                task=escape(progress.name_during),
                progress=100.0 * progress.progress,
                eta=timestamp2timedelta(progress.eta()),
            )

        if progress.max_steps is None:
            update = self._update_unknown_steps(progress)
        else:
            update = self._update_known_steps(progress)

        if progress.finished:
            finish = """
                fill.style.animation = 'none';
                fill.style.backgroundImage = 'none';
            """
        else:
            finish = ""

        return Javascript(
            f"""
              (function () {{
                  var root = document.getElementById('{self._uuid}');
                  var text = root.getElementsByClassName('pb-text')[0];
                  var fill = root.getElementsByClassName('pb-fill')[0];

                  text.innerHTML = '{text}';
                  {update}
                  {finish}
              }})();
        """
        )

    def _update_known_steps(self, progress):
        return """
            if ({progress} > 0.) {{
                fill.style.transition = 'width 0.1s linear';
            }} else {{
                fill.style.transition = 'none';
            }}

            fill.style.width = '{progress}%';
            fill.style.animation = 'none';
            fill.style.backgroundImage = 'none'
        """.format(
            progress=100.0 * progress.progress
        )

    def _update_unknown_steps(self, progress):
        return """
            fill.style.width = '100%';
            fill.style.animation = 'pb-fill-anim 2s linear infinite';
            fill.style.backgroundSize = '100px 100%';
            fill.style.backgroundImage = 'repeating-linear-gradient(' +
                '90deg, #bdd2e6, #edf2f8 40%, #bdd2e6 80%, #bdd2e6)';
        """


class VdomOrHtmlProgressBar(ProgressBar):  # pragma: no cover
    """Progress bar using the VDOM or HTML progress bar.

    This progress bar will transmit both representations as part of a MIME
    bundle and it is up to the Jupyter client to pick the preferred version.
    Usually this will be the VDOM if supported, and the HMTL version where VDOM
    is not supported.

    .. versionadded: 3.0.0
    """

    def __init__(self):
        super().__init__()
        self._handle = None
        self._vdom = VdomProgressBar()
        self._html = HtmlProgressBar()

    def update(self, progress):
        self._vdom.progress = progress
        if self._handle is None:
            display(self._get_initial_bundle(progress), raw=True)
            self._handle = display(
                self._get_update_bundle(progress), raw=True, display_id=True
            )
        self._handle.update(self._get_update_bundle(progress), raw=True)

    def _get_initial_bundle(self, progress):
        return {
            "application/vdom.v1+json": {"tagName": "div", "attributes": {}},
            "text/html": self._html._HtmlBase(self._html._uuid)._repr_html_(),
            "text/plain": repr(self._html._HtmlBase(self._html._uuid)),
        }

    def _get_update_bundle(self, progress):
        bundle = self._vdom._repr_mimebundle_([], [])
        bundle["text/html"] = (
            "<script>"
            + self._html._js_update(progress)._repr_javascript_()
            + "</script>"
        )
        return bundle


class IPython5ProgressBar(ProgressBar):  # pragma: no cover
    """ProgressBar for IPython>=5 environments.

    Provides a VDOM/HTML representation, except for in a pure terminal IPython
    (i.e. not an IPython kernel that was connected to via ZMQ), where a
    ASCII progress bar will be used.

    Note that some Jupyter environments (like qtconsole) will try to use the
    VDOM/HTML version, but do not support HTML and will show a warning instead
    of an actual progress bar.
    """

    def __init__(self):
        super().__init__()

        class Displayable:
            def __init__(self):
                self.display_requested = False

            def _ipython_display_(self):
                self.display_requested = True

        d = Displayable()
        display(d, exclude=["text/plain"])

        if d.display_requested:
            self._progress_bar = VdomOrHtmlProgressBar()
        else:
            self._progress_bar = TerminalProgressBar()

    def update(self, progress):
        self._progress_bar.update(progress)


class WriteProgressToFile(ProgressBar):
    """Writes progress to a file.

    This is useful for remotely and intermittently monitoring progress.
    Note that this file will be overwritten on each update of the progress!

    Parameters
    ----------
    filename : str
        Path to the file to write the progress to.
    """

    def __init__(self, filename):
        self.filename = filename
        super().__init__()

    def update(self, progress):
        if progress.finished:
            text = (
                "{} finished in {}.".format(  # pylint: disable=consider-using-f-string
                    progress.name_after, timestamp2timedelta(progress.elapsed_seconds())
                )
            )
        else:
            text = "{progress:.0f}%, ETA: {eta}".format(  # pylint: disable=consider-using-f-string
                progress=100 * progress.progress,
                eta=timestamp2timedelta(progress.eta()),
            )

        with open(self.filename, "w", encoding="utf-8") as f:
            f.write(text + os.linesep)


class AutoProgressBar(ProgressBar):
    """Suppresses the progress bar unless the ETA exceeds a threshold.

    Parameters
    ----------
    delegate : ProgressBar
        The actual progress bar to display, if ETA is high enough.
    min_eta : float, optional
        The minimum ETA threshold for displaying the progress bar.
    """

    def __init__(self, delegate, min_eta=1.0):
        self.delegate = delegate

        super().__init__()

        self.min_eta = min_eta
        self._visible = False

    def update(self, progress):
        min_delay = progress.time_start + 0.1
        long_eta = (
            progress.elapsed_seconds() + progress.eta() > self.min_eta
            and min_delay < time.time()
        )
        if self._visible:
            self.delegate.update(progress)
        elif long_eta or progress.finished:
            self._visible = True
            self.delegate.update(progress)

    def close(self):
        self.delegate.close()


class ProgressTracker:
    """Tracks the progress of some process with a progress bar.

    Parameters
    ----------
    progress_bar : ProgressBar or bool or None
        The progress bar to display the progress (or True to use the default
        progress bar, False/None to disable progress bar).
    total_progress : int
        Maximum number of steps of the process.
    update_interval : float, optional
        Time to wait (in seconds) between updates to progress bar display.
    """

    def __init__(self, progress_bar, total_progress, update_interval=0.1):
        self.progress_bar = to_progressbar(progress_bar)
        self.total_progress = total_progress
        self.update_interval = update_interval
        self.update_thread = threading.Thread(target=self.update_loop)
        self.update_thread.daemon = True
        self._closing = False
        self.sub_progress = None

    def next_stage(self, name_during="", name_after=None, max_steps=None):
        """Begin tracking progress of a new stage.

        Parameters
        ----------
        max_steps : int, optional
            The total number of calculation steps of the process.
        name_during : str, optional
            Short description of the task to be used while it is running.
        name_after : str, optional
            Short description of the task to be used after it has
            finished. Defaults to *name_during*.
        """
        if self.sub_progress is not None:
            self.total_progress.step()
        self.sub_progress = Progress(name_during, name_after, max_steps)
        return self.sub_progress

    def __enter__(self):
        self._closing = False
        self.total_progress.__enter__()
        if not isinstance(self.progress_bar, NoProgressBar):
            self.update_thread.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._closing = True
        self.total_progress.__exit__(exc_type, exc_value, traceback)
        if not isinstance(self.progress_bar, NoProgressBar):
            self.update_thread.join()
        self.progress_bar.update(self.total_progress)
        self.progress_bar.close()

    def update_loop(self):
        """Update the progress bar display (will run in a separate thread)."""

        while not self._closing:
            if self.sub_progress is not None and not self.sub_progress.finished:
                self.progress_bar.update(self.sub_progress)
            else:
                self.progress_bar.update(self.total_progress)
            time.sleep(self.update_interval)


def get_default_progressbar():
    """The default progress bar to use depending on the execution environment.

    Returns
    -------
    ``ProgressBar``
    """
    try:
        pbar = rc["progress"].getboolean("progress_bar")
        if pbar:
            pbar = "auto"
        else:
            pbar = "none"
    except ValueError:
        pbar = rc["progress"]["progress_bar"]

    if pbar.lower() == "auto":
        if get_ipython() is not None and check_ipy_version((5, 0)):  # pragma: no cover
            return AutoProgressBar(IPython5ProgressBar())
        else:
            return AutoProgressBar(TerminalProgressBar())
    if pbar.lower() == "none":
        return NoProgressBar()

    return _load_class(pbar)()


def to_progressbar(progress_bar):
    """Converts to a ``ProgressBar`` instance.

    Parameters
    ----------
    progress_bar : None, bool, or ProgressBar
        Object to be converted to a ``ProgressBar``.

    Returns
    -------
    ProgressBar
        Return ``progress_bar`` if it is already a progress bar, the default
        progress bar if ``progress_bar`` is ``True``, and ``NoProgressBar`` if
        it is ``None`` or ``False``.
    """
    if progress_bar is False or progress_bar is None:
        progress_bar = NoProgressBar()
    if progress_bar is True:
        progress_bar = get_default_progressbar()
    return progress_bar



"""
Extra functions to extend the capabilities of Numpy.
"""
import logging
import warnings
from collections.abc import Iterable
from distutils.version import LooseVersion

import numpy as np

from ..exceptions import ValidationError

logger = logging.getLogger(__name__)
try:
    import scipy.sparse as scipy_sparse

    def is_spmatrix(obj):
        """Check if ``obj`` is a sparse matrix."""
        return isinstance(obj, scipy_sparse.spmatrix)

except ImportError as e:
    logger.info("Could not import scipy.sparse:\n%s", str(e))
    scipy_sparse = None

    def is_spmatrix(obj):
        """Check if ``obj`` is a sparse matrix."""
        return False


maxseed = np.iinfo(np.uint32).max
maxint = np.iinfo(np.int32).max


# numpy 1.17 introduced a slowdown to clip, so
# use nengo.utils.numpy.clip instead of np.clip
# This has persisted through 1.19 at least
clip = np.core.umath.clip if LooseVersion(np.__version__) >= "1.17.0" else np.clip


def is_integer(obj):
    """Check if ``obj`` is an integer type."""
    return isinstance(obj, (int, np.integer))


def is_iterable(obj):
    """Check if ``obj`` is an iterable."""
    if isinstance(obj, np.ndarray):
        return obj.ndim > 0  # 0-d arrays give error if iterated over
    else:
        return isinstance(obj, Iterable)


def is_number(obj, check_complex=False):
    """Check if ``obj`` is a numeric type."""
    types = (float, complex, np.number) if check_complex else (float, np.floating)
    return is_integer(obj) or isinstance(obj, types)


def is_array(obj):
    """Check if ``obj`` is a numpy array."""
    # np.generic allows us to return true for scalars as well as true arrays
    return isinstance(obj, (np.ndarray, np.generic))


def is_array_like(obj):
    """Check if ``obj`` is an array like object."""
    # While it's possible that there are some iterables other than list/tuple
    # that can be made into arrays, it's very likely that those arrays
    # will have dtype=object, which is likely to cause unexpected issues.
    return is_array(obj) or is_number(obj) or isinstance(obj, (list, tuple))


def compare(a, b):
    """Return -1/0/1 if a is less/equal/greater than b."""

    return 0 if a == b else 1 if a > b else -1 if a < b else None


def as_shape(x, min_dim=0):
    """Return a tuple if ``x`` is iterable or ``(x,)`` if ``x`` is integer."""
    if is_iterable(x):
        shape = tuple(x)
    elif is_integer(x):
        shape = (x,)
    else:
        raise ValueError(f"{x!r} cannot be safely converted to a shape")

    if len(shape) < min_dim:
        shape = tuple([1] * (min_dim - len(shape))) + shape

    return shape


def broadcast_shape(shape, length):
    """Pad a shape with ones following standard Numpy broadcasting."""
    n = len(shape)
    if n < length:
        return tuple([1] * (length - n) + list(shape))
    else:
        return shape


def array(x, dims=None, min_dims=0, readonly=False, **kwargs):
    """Create numpy array with some extra configuration.

    This is a wrapper around ``np.array``.

    Unlike ``np.array``, the additional single-dimensional indices added by
    ``dims`` or ``min_dims`` will appear at the *end* of the shape (for example,
    ``array([1, 2, 3], dims=4).shape == (3, 1, 1, 1)``).

    Parameters
    ----------
    dims : int or None
        If not ``None``, force the output array to have exactly this many indices.
        If the input has more than this number of indices, this throws an error.
    min_dims : int
        Force the output array to have at least this many indices
        (ignored if ``dims is not None``).
    readonly : bool
        Make the output array read-only.
    **kwargs
        Additional keyword arguments to pass to ``np.array``.
    """

    y = np.array(x, **kwargs)
    dims = max(min_dims, y.ndim) if dims is None else dims

    if y.ndim < dims:
        shape = np.ones(dims, dtype="int")
        shape[: y.ndim] = y.shape
        y.shape = shape
    elif y.ndim > dims:
        raise ValidationError(
            f"Input cannot be cast to array with {dims} dimensions", attr="dims"
        )

    if readonly:
        y.flags.writeable = False

    return y


def _array_hash(a, n=100):
    if not isinstance(a, np.ndarray):
        return hash(a)

    if a.size < n:
        # hash all elements
        v = a.view()
        v.setflags(write=False)
        return hash(v.data.tobytes())
    else:
        # pick random elements to hash
        rng = np.random.RandomState(a.size)
        inds = tuple(rng.randint(0, a.shape[i], size=n) for i in range(a.ndim))
        v = a[inds]
        v.setflags(write=False)
        return hash(v.data.tobytes())


def array_hash(a, n=100):
    """Simple fast array hash function.

    For arrays with size larger than ``n``, pick ``n`` elements at random
    to hash. This strategy should work well for dense arrays, but for
    sparse arrays (those with few non-zero elements) it is more likely to
    give hash collisions.

    For sparse matrices, we apply the same logic on the underlying elements
    (data, indices) of the sparse matrix.
    """
    if scipy_sparse and isinstance(a, scipy_sparse.spmatrix):
        if isinstance(
            a,
            (scipy_sparse.csr_matrix, scipy_sparse.csc_matrix, scipy_sparse.bsr_matrix),
        ):
            return hash(
                (
                    _array_hash(a.data, n=n),
                    _array_hash(a.indices, n=n),
                    _array_hash(a.indptr, n=n),
                )
            )
        else:
            if not isinstance(a, scipy_sparse.coo_matrix):
                a = a.tocoo()
            return hash(
                (
                    _array_hash(a.data, n=n),
                    _array_hash(a.row, n=n),
                    _array_hash(a.col, n=n),
                )
            )

    return _array_hash(a, n=n)


def array_offset(x):
    """Get offset of array data from base data in bytes."""
    if x.base is None:
        return 0

    base_start = x.base.__array_interface__["data"][0]
    start = x.__array_interface__["data"][0]
    return start - base_start


def norm(x, axis=None, keepdims=False):
    """Compute the Euclidean norm.

    Parameters
    ----------
    x : array_like
        Array to compute the norm over.
    axis : None or int or tuple of ints, optional
        Axis or axes to sum across. ``None`` sums all axes. See ``np.sum``.
    keepdims : bool, optional
        If True, the reduced axes are left in the result. See ``np.sum`` in
        newer versions of Numpy (>= 1.7).
    """
    x = np.asarray(x)
    return np.sqrt(np.sum(x ** 2, axis=axis, keepdims=keepdims))


def meshgrid_nd(*args):
    """Multidimensional meshgrid."""
    args = [np.asarray(a) for a in args]
    s = len(args) * (1,)
    return np.broadcast_arrays(
        *(a.reshape(s[:i] + (-1,) + s[i + 1 :]) for i, a in enumerate(args))
    )


def rms(x, axis=None, keepdims=False):
    """Compute the root-mean-square amplitude.

    Parameters
    ----------
    x : array_like
        Array to compute RMS amplitude over.
    axis : None or int or tuple of ints, optional
        Axis or axes to sum across. ``None`` sums all axes. See ``np.sum``.
    keepdims : bool, optional
        If True, the reduced axes are left in the result. See ``np.sum`` in
        newer versions of Numpy (>= 1.7).
    """
    x = np.asarray(x)
    return np.sqrt(np.mean(x ** 2, axis=axis, keepdims=keepdims))


def rmse(x, y, axis=None, keepdims=False):  # pragma: no cover
    """Compute the root-mean-square error.

    Equivalent to rms(x - y, axis=axis, keepdims=keepdims).

    Parameters
    ----------
    x, y : array_like
        Arrays to compute RMS amplitude over.
    axis : None or int or tuple of ints, optional
        Axis or axes to sum across. ``None`` sums all axes. See ``np.sum``.
    keepdims : bool, optional
        If True, the reduced axes are left in the result. See ``np.sum`` in
        newer versions of Numpy (>= 1.7).
    """
    warnings.warn(
        "The 'rmse' function is deprecated and will be removed in a future "
        "version. Please use ``rms(x - y)`` instead.",
        DeprecationWarning,
    )
    x, y = np.asarray(x), np.asarray(y)
    return rms(x - y, axis=axis, keepdims=keepdims)


def nrmse(a, b, axis=None, keepdims=False):
    """Compute the root-mean-square (RMS) error normalized by the RMS of ``b``.

    Equivalent to ``rms(a - b, **kwargs) / rms(b, **kwargs)``

    Parameters
    ----------
    a, b : array_like
        Arrays to compute RMS error over, normalized by the rms amplitude of ``b``.
    axis : None or int or tuple of ints, optional
        Axis or axes to sum across. ``None`` sums all axes. See ``np.sum``.
    keepdims : bool, optional
        If True, the reduced axes are left in the result. See ``np.sum`` in
        newer versions of Numpy (>= 1.7).
    """
    a, b = np.asarray(a), np.asarray(b)
    return rms(a - b, axis=axis, keepdims=keepdims) / rms(
        b, axis=axis, keepdims=keepdims
    )


if hasattr(np.fft, "rfftfreq"):
    rfftfreq = np.fft.rfftfreq
else:

    def rfftfreq(n, d=1.0):
        return np.abs(np.fft.fftfreq(n=n, d=d)[: n // 2 + 1])



"""
These are borrowed from SciPy and used under their license:

Copyright (c) 2001, 2002 Enthought, Inc.
All rights reserved.

Copyright (c) 2003-2012 SciPy Developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  a. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
  b. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  c. Neither the name of Enthought nor the names of the SciPy Developers
     may be used to endorse or promote products derived from this software
     without specific prior written permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.
"""
# From scipy/sparse/sputils.py v0.18.0

from __future__ import absolute_import, division

import numpy as np


def isscalarlike(x):
    """Is x either a scalar, an array scalar, or a 0-dim array?"""
    return np.isscalar(x) or (isdense(x) and x.ndim == 0)


def isintlike(x):
    """Is x appropriate as an index into a sparse matrix? Returns True
    if it can be cast safely to a machine int.
    """
    if issequence(x):
        return False
    try:
        return bool(int(x) == x)
    except (TypeError, ValueError):
        return False


def isshape(x):
    """Is x a valid 2-tuple of dimensions?
    """
    try:
        # Assume it's a tuple of matrix dimensions (M, N)
        (M, N) = x
    except:
        return False
    else:
        if isintlike(M) and isintlike(N):
            if np.ndim(M) == 0 and np.ndim(N) == 0:
                return True
        return False


def issequence(t):
    return ((isinstance(t, (list, tuple)) and
            (len(t) == 0 or np.isscalar(t[0]))) or
            (isinstance(t, np.ndarray) and (t.ndim == 1)))


def ismatrix(t):
    return ((isinstance(t, (list, tuple)) and
             len(t) > 0 and issequence(t[0])) or
            (isinstance(t, np.ndarray) and t.ndim == 2))


def isdense(x):
    return isinstance(x, np.ndarray)



"""Classes concerned with solving for decoders or full weight matrices.

.. inheritance-diagram:: nengo.solvers
   :parts: 1
   :top-classes: nengo.solvers.Solver, nengo.solvers.SolverParam

"""

import time

import numpy as np

import nengo.utils.least_squares_solvers as lstsq
from nengo.params import BoolParam, FrozenObject, NdarrayParam, NumberParam, Parameter
from nengo.utils.least_squares_solvers import (
    LeastSquaresSolverParam,
    format_system,
    rmses,
)


class Solver(FrozenObject):
    """Decoder or weight solver.

    A solver can have the ``weights`` parameter equal to ``True`` or ``False``.

    Weight solvers are used to form neuron-to-neuron weight matrices.
    They can be compositional or non-compositional. Non-compositional
    solvers must operate on the whole neuron-to-neuron weight matrix
    (i.e., each target is a separate postsynaptic current, without the bias
    term), while compositional solvers operate in the decoded state-space
    (i.e., each target is a dimension in state-space). Compositional solvers
    then combine the returned ``X`` with the transform and/or encoders to
    generate the full weight matrix.

    For a solver to be compositional, the following property must be true::

        X = solver(A, Y)  if and only if  L(X) = solver(A, L(Y))

    where ``L`` is some arbitrary linear operator (i.e., the transform and/or
    encoders for the postsynaptic population). This property can then be
    leveraged by the backend for efficiency. See the solver's
    ``compositional`` class attribute to determine if it is compositional.

    Non-weight solvers always operate in the decoded state-space regardless of
    whether they are compositional or non-compositional.
    """

    compositional = True

    weights = BoolParam("weights")

    def __init__(self, weights=False):
        super().__init__()
        self.weights = weights

    def __call__(self, A, Y, rng=np.random):
        """Call the solver.

        .. note:: ``n_targets`` is ``dimensions`` if ``solver.weights`` is ``False``
                  and ``post.n_neurons`` if ``solver.weights`` is ``True``.

        Parameters
        ----------
        A : (n_eval_points, n_neurons) array_like
            Matrix of the neurons' activities at the evaluation points.
        Y : (n_eval_points, n_targets) array_like
            Matrix of target values at the evaluation points.
        rng : `numpy.random.RandomState`, optional
            A random number generator to use as required.

        Returns
        -------
        X : (n_neurons, n_targets) array_like
            Matrix of weights used to map activities onto targets.
            A typical solver will approximate ``dot(A, X) ~= Y`` subject to
            some constraints on ``X``.
        info : dict
            A dictionary of information about the solver. All dictionaries have
            an ``'rmses'`` key that contains RMS errors of the solve (one per
            target). Other keys are unique to particular solvers.

        """
        raise NotImplementedError("Solvers must implement '__call__'")


class SolverParam(Parameter):
    """A parameter in which the value is a `.Solver` instance."""

    def coerce(self, instance, solver):  # pylint: disable=arguments-renamed
        self.check_type(instance, solver, Solver)
        return super().coerce(instance, solver)


class Lstsq(Solver):
    """Unregularized least-squares solver.

    Parameters
    ----------
    weights : bool, optional
        If False, solve for decoders. If True, solve for weights.
    rcond : float, optional
        Cut-off ratio for small singular values (see `numpy.linalg.lstsq`).

    Attributes
    ----------
    rcond : float
        Cut-off ratio for small singular values (see `numpy.linalg.lstsq`).
    weights : bool
        If False, solve for decoders. If True, solve for weights.
    """

    rcond = NumberParam("noise", low=0)

    def __init__(self, weights=False, rcond=0.01):
        super().__init__(weights=weights)
        self.rcond = rcond

    def __call__(self, A, Y, rng=np.random):
        tstart = time.time()
        X, residuals2, rank, s = np.linalg.lstsq(A, Y, rcond=self.rcond)
        t = time.time() - tstart
        return (
            X,
            {
                "rmses": rmses(A, X, Y),
                "residuals": np.sqrt(residuals2),
                "rank": rank,
                "singular_values": s,
                "time": t,
            },
        )


def _add_noise_param_docs(cls):
    cls.__doc__ += """

    Parameters
    ----------
    weights : bool, optional
        If False, solve for decoders. If True, solve for weights.
    noise : float, optional
        Amount of noise, as a fraction of the neuron activity.
    solver : `.LeastSquaresSolver`, optional
        Subsolver to use for solving the least squares problem.

    Attributes
    ----------
    noise : float
        Amount of noise, as a fraction of the neuron activity.
    solver : `.LeastSquaresSolver`
        Subsolver to use for solving the least squares problem.
    weights : bool
        If False, solve for decoders. If True, solve for weights.
    """
    return cls


@_add_noise_param_docs
class LstsqNoise(Solver):
    """Least-squares solver with additive Gaussian white noise."""

    noise = NumberParam("noise", low=0)
    solver = LeastSquaresSolverParam("solver")

    def __init__(self, weights=False, noise=0.1, solver=lstsq.Cholesky()):
        super().__init__(weights=weights)
        self.noise = noise
        self.solver = solver

    def __call__(self, A, Y, rng=np.random):
        tstart = time.time()
        sigma = self.noise * np.amax(np.abs(A))
        A = A + rng.normal(scale=sigma, size=A.shape)
        X, info = self.solver(A, Y, 0, rng=rng)
        info["time"] = time.time() - tstart
        return X, info


@_add_noise_param_docs
class LstsqMultNoise(LstsqNoise):
    """Least-squares solver with multiplicative white noise."""

    def __call__(self, A, Y, rng=np.random):
        tstart = time.time()
        A = A + rng.normal(scale=self.noise, size=A.shape) * A
        X, info = self.solver(A, Y, 0, rng=rng)
        info["time"] = time.time() - tstart
        return X, info


def _add_l2_param_docs(cls):
    cls.__doc__ += """

    Parameters
    ----------
    weights : bool, optional
        If False, solve for decoders. If True, solve for weights.
    reg : float, optional
        Amount of regularization, as a fraction of the neuron activity.
    solver : `.LeastSquaresSolver`, optional
        Subsolver to use for solving the least squares problem.

    Attributes
    ----------
    reg : float
        Amount of regularization, as a fraction of the neuron activity.
    solver : `.LeastSquaresSolver`
        Subsolver to use for solving the least squares problem.
    weights : bool
        If False, solve for decoders. If True, solve for weights.
    """
    return cls


@_add_l2_param_docs
class LstsqL2(Solver):
    """Least-squares solver with L2 regularization."""

    reg = NumberParam("reg", low=0)
    solver = LeastSquaresSolverParam("solver")

    def __init__(self, weights=False, reg=0.1, solver=lstsq.Cholesky()):
        super().__init__(weights=weights)
        self.reg = reg
        self.solver = solver

    def __call__(self, A, Y, rng=np.random):
        tstart = time.time()
        sigma = self.reg * A.max()
        X, info = self.solver(A, Y, sigma, rng=rng)
        info["time"] = time.time() - tstart
        return X, info


@_add_l2_param_docs
class LstsqL2nz(LstsqL2):
    """Least-squares solver with L2 regularization on non-zero components."""

    def __call__(self, A, Y, rng=np.random):
        tstart = time.time()
        # Compute the equivalent noise standard deviation. This equals the
        # base amplitude (noise_amp times the overall max activation) times
        # the square-root of the fraction of non-zero components.
        sigma = (self.reg * np.amax(np.abs(A))) * np.sqrt((np.abs(A) > 0).mean(axis=0))

        # sigma == 0 means the neuron is never active, so won't be used, but
        # we have to make sigma != 0 for numeric reasons.
        sigma[sigma == 0] = sigma.max()

        X, info = self.solver(A, Y, sigma, rng=rng)
        info["time"] = time.time() - tstart
        return X, info


class LstsqL1(Solver):
    """Least-squares solver with L1 and L2 regularization (elastic net).

    This method is well suited for creating sparse decoders or weight matrices.

    .. note:: Requires `scikit-learn <https://scikit-learn.org/stable/>`_.

    Parameters
    ----------
    weights : bool, optional
        If False, solve for decoders. If True, solve for weights.
    l1 : float, optional
        Amount of L1 regularization.
    l2 : float, optional
        Amount of L2 regularization.
    max_iter : int, optional
        Maximum number of iterations for the underlying elastic net.

    Attributes
    ----------
    l1 : float
        Amount of L1 regularization.
    l2 : float
        Amount of L2 regularization.
    weights : bool
        If False, solve for decoders. If True, solve for weights.
    max_iter : int
        Maximum number of iterations for the underlying elastic net.
    """

    compositional = False

    l1 = NumberParam("l1", low=0)
    l2 = NumberParam("l2", low=0)

    def __init__(self, weights=False, l1=1e-4, l2=1e-6, max_iter=1000):
        # import to check existence
        import sklearn.linear_model  # pylint: disable=import-outside-toplevel

        assert sklearn.linear_model
        super().__init__(weights=weights)
        self.l1 = l1
        self.l2 = l2
        self.max_iter = max_iter

    def __call__(self, A, Y, rng=np.random):
        import sklearn.linear_model  # pylint: disable=import-outside-toplevel

        tstart = time.time()
        Y = np.array(Y)  # copy since 'fit' may modify Y

        # TODO: play around with regularization constants (I just guessed).
        #   Do we need to scale regularization by number of neurons, to get
        #   same level of sparsity? esp. with weights? Currently, setting
        #   l1=1e-3 works well with weights when connecting 1D populations
        #   with 100 neurons each.
        a = self.l1 * A.max()  # L1 regularization
        b = self.l2 * A.max() ** 2  # L2 regularization
        alpha = a + b
        l1_ratio = a / (a + b)

        # --- solve least-squares A * X = Y
        model = sklearn.linear_model.ElasticNet(
            alpha=alpha, l1_ratio=l1_ratio, fit_intercept=False, max_iter=self.max_iter
        )
        model.fit(A, Y)
        X = model.coef_.T
        X.shape = (A.shape[1], Y.shape[1]) if Y.ndim > 1 else (A.shape[1],)
        t = time.time() - tstart
        infos = {"rmses": rmses(A, X, Y), "time": t}
        return X, infos


class LstsqDrop(Solver):
    """Find sparser decoders/weights by dropping small values.

    This solver first solves for coefficients (decoders/weights) with
    L2 regularization, drops those nearest to zero, and retrains remaining.

    Parameters
    ----------
    weights : bool, optional
        If False, solve for decoders. If True, solve for weights.
    drop : float, optional
        Fraction of decoders or weights to set to zero.
    solver1 : Solver, optional
        Solver for finding the initial decoders.
    solver2 : Solver, optional
        Used for re-solving for the decoders after dropout.

    Attributes
    ----------
    drop : float
        Fraction of decoders or weights to set to zero.
    solver1 : Solver
        Solver for finding the initial decoders.
    solver2 : Solver
        Used for re-solving for the decoders after dropout.
    weights : bool
        If False, solve for decoders. If True, solve for weights.
    """

    compositional = False

    drop = NumberParam("drop", low=0, high=1)
    solver1 = SolverParam("solver1")
    solver2 = SolverParam("solver2")

    def __init__(
        self,
        weights=False,
        drop=0.25,
        solver1=LstsqL2(reg=0.001),
        solver2=LstsqL2(reg=0.1),
    ):
        super().__init__(weights=weights)
        self.drop = drop
        self.solver1 = solver1
        self.solver2 = solver2

    def __call__(self, A, Y, rng=np.random):
        tstart = time.time()
        Y, _, _, _, matrix_in = format_system(A, Y)

        # solve for coefficients using standard solver
        X, info0 = self.solver1(A, Y, rng=rng)

        # drop weights close to zero, based on `drop` ratio
        Xabs = np.sort(np.abs(X.flat))
        threshold = Xabs[int(np.round(self.drop * Xabs.size))]
        X[np.abs(X) < threshold] = 0

        # retrain nonzero weights
        info1s = []
        for i in range(X.shape[1]):
            info1 = None
            nonzero = X[:, i] != 0
            if nonzero.sum() > 0:
                X[nonzero, i], info1 = self.solver2(A[:, nonzero], Y[:, i], rng=rng)
            info1s.append(info1)

        t = time.time() - tstart
        info = {"rmses": rmses(A, X, Y), "info0": info0, "info1s": info1s, "time": t}
        return X if matrix_in or X.shape[1] > 1 else X.ravel(), info


def _add_nnls_param_docs(l2=False):
    reg_attr = """
    reg : float
        Amount of regularization, as a fraction of the neuron activity.\
    """
    reg_param = """
    reg : float, optional
        Amount of regularization, as a fraction of the neuron activity.\
    """

    docstring = f"""
    .. note:: Requires
              `SciPy <https://docs.scipy.org/doc/scipy/reference/>`_.

    Parameters
    ----------
    weights : bool, optional
        If False, solve for decoders. If True, solve for weights.\
{reg_param if l2 else ""}

    Attributes
    ----------{reg_attr if l2 else ""}
    weights : bool
        If False, solve for decoders. If True, solve for weights.
    """

    def _actually_add_nnls_param_docs(cls):
        cls.__doc__ += docstring
        return cls

    return _actually_add_nnls_param_docs


@_add_nnls_param_docs(l2=False)
class Nnls(Solver):
    """Non-negative least-squares solver without regularization.

    Similar to `.Lstsq`, except the output values are non-negative.

    If solving for non-negative **weights**, it is important that the
    intercepts of the post-population are also non-negative, since neurons with
    negative intercepts will never be silent, affecting output accuracy.
    """

    compositional = False

    def __init__(self, weights=False):
        # import here too to throw error early
        import scipy.optimize  # pylint: disable=import-outside-toplevel

        assert scipy.optimize
        super().__init__(weights=weights)

    def __call__(self, A, Y, rng=np.random):
        import scipy.optimize  # pylint: disable=import-outside-toplevel

        tstart = time.time()
        Y, _, n, _, matrix_in = format_system(A, Y)
        d = Y.shape[1]

        X = np.zeros((n, d))
        residuals = np.zeros(d)
        for i in range(d):
            X[:, i], residuals[i] = scipy.optimize.nnls(A, Y[:, i])

        t = time.time() - tstart
        info = {"rmses": rmses(A, X, Y), "residuals": residuals, "time": t}
        return X if matrix_in or X.shape[1] > 1 else X.ravel(), info


@_add_nnls_param_docs(l2=True)
class NnlsL2(Nnls):
    """Non-negative least-squares solver with L2 regularization.

    Similar to `.LstsqL2`, except the output values are non-negative.

    If solving for non-negative **weights**, it is important that the
    intercepts of the post-population are also non-negative, since neurons with
    negative intercepts will never be silent, affecting output accuracy.
    """

    reg = NumberParam("reg", low=0)

    def __init__(self, weights=False, reg=0.1):
        super().__init__(weights=weights)
        self.reg = reg

    def _solve(self, A, Y, sigma=0.0):
        import scipy.optimize  # pylint: disable=import-outside-toplevel

        tstart = time.time()
        Y, _, n, _, matrix_in = format_system(A, Y)
        d = Y.shape[1]

        # form Gram matrix so we can add regularization
        GA = np.dot(A.T, A)
        np.fill_diagonal(GA, GA.diagonal() + A.shape[0] * sigma ** 2)
        GY = np.dot(A.T, np.maximum(Y, 0))
        # ^ TODO: why is it better if we clip Y to be positive here?

        X = np.zeros((n, d))
        residuals = np.zeros(d)
        for i in range(d):
            X[:, i], residuals[i] = scipy.optimize.nnls(GA, GY[:, i])

        t = time.time() - tstart
        info = {"rmses": rmses(A, X, Y), "residuals": residuals, "time": t}
        return X if matrix_in or X.shape[1] > 1 else X.ravel(), info

    def __call__(self, A, Y, rng=np.random):
        return self._solve(A, Y, sigma=self.reg * A.max())


@_add_nnls_param_docs(l2=True)
class NnlsL2nz(NnlsL2):
    """Non-negative least-squares with L2 regularization on nonzero components.

    Similar to `.LstsqL2nz`, except the output values are non-negative.

    If solving for non-negative **weights**, it is important that the
    intercepts of the post-population are also non-negative, since neurons with
    negative intercepts will never be silent, affecting output accuracy.
    """

    def __call__(self, A, Y, rng=np.random):
        sigma = (self.reg * A.max()) * np.sqrt((A > 0).mean(axis=0))
        sigma[sigma == 0] = 1
        return self._solve(A, Y, sigma=sigma)


class NoSolver(Solver):
    """Manually pass in weights, bypassing the decoder solver.

    Parameters
    ----------
    values : (n_neurons, size_out) array_like, optional
        The array of decoders to use.
        ``size_out`` is the dimensionality of the decoded signal (determined
        by the connection function).
        If ``None``, which is the default, the solver will return an
        appropriately sized array of zeros.
    weights : bool, optional
        If False, connection will use factored weights (decoders from this
        solver, transform, and encoders).
        If True, connection will use a full weight matrix (created by
        linearly combining decoder, transform, and encoders).

    Attributes
    ----------
    values : (n_neurons, size_out) array_like, optional
        The array of decoders to use.
        ``size_out`` is the dimensionality of the decoded signal (determined
        by the connection function).
        If ``None``, which is the default, the solver will return an
        appropriately sized array of zeros.
    weights : bool, optional
        If False, connection will use factored weights (decoders from this
        solver, transform, and encoders).
        If True, connection will use a full weight matrix (created by
        linearly combining decoder, transform, and encoders).
    """

    compositional = True

    values = NdarrayParam("values", optional=True, shape=("*", "*"))

    def __init__(self, values=None, weights=False):
        super().__init__(weights=weights)
        self.values = values

    def __call__(self, A, Y, rng=None):
        if self.values is None:
            n_neurons = np.asarray(A).shape[1]

            return np.zeros((n_neurons, np.asarray(Y).shape[1])), {}

        return self.values, {}
